import base64
import copy
import json
import uuid
from typing import AsyncIterator, List, Optional

import httpx
import pydantic_core
from google.genai import Client, errors
from google.genai.types import (
    FunctionCallingConfig,
    FunctionCallingConfigMode,
    GenerateContentResponse,
    HttpOptions,
    ThinkingConfig,
    ToolConfig,
)

from letta.constants import NON_USER_MSG_PREFIX
from letta.errors import (
    ContextWindowExceededError,
    ErrorCode,
    LLMAuthenticationError,
    LLMBadRequestError,
    LLMConnectionError,
    LLMInsufficientCreditsError,
    LLMNotFoundError,
    LLMPermissionDeniedError,
    LLMRateLimitError,
    LLMServerError,
    LLMTimeoutError,
    LLMUnprocessableEntityError,
)
from letta.helpers.datetime_helpers import get_utc_time_int
from letta.helpers.json_helpers import json_dumps, json_loads, sanitize_unicode_surrogates
from letta.llm_api.error_utils import is_insufficient_credits_message
from letta.llm_api.llm_client_base import LLMClientBase
from letta.local_llm.json_parser import clean_json_string_extra_backslash
from letta.log import get_logger
from letta.otel.tracing import trace_method
from letta.schemas.agent import AgentType
from letta.schemas.enums import ProviderCategory
from letta.schemas.llm_config import LLMConfig
from letta.schemas.message import Message as PydanticMessage
from letta.schemas.openai.chat_completion_request import Tool, Tool as OpenAITool
from letta.schemas.openai.chat_completion_response import ChatCompletionResponse, Choice, FunctionCall, Message, ToolCall, UsageStatistics
from letta.schemas.usage import LettaUsageStatistics
from letta.settings import model_settings, settings
from letta.utils import get_tool_call_id

logger = get_logger(__name__)


class GoogleVertexClient(LLMClientBase):
    MAX_RETRIES = model_settings.gemini_max_retries
    provider_label = "Google Vertex"

    def _get_client(self, llm_config: Optional[LLMConfig] = None):
        timeout_ms = int(settings.llm_request_timeout_seconds * 1000)
        if llm_config:
            api_key, _, _ = self.get_byok_overrides(llm_config)
            if api_key:
                return Client(
                    api_key=api_key,
                    http_options=HttpOptions(timeout=timeout_ms),
                )
        return Client(
            vertexai=True,
            project=model_settings.google_cloud_project,
            location=model_settings.google_cloud_location,
            http_options=HttpOptions(api_version="v1", timeout=timeout_ms),
        )

    async def _get_client_async(self, llm_config: Optional[LLMConfig] = None):
        timeout_ms = int(settings.llm_request_timeout_seconds * 1000)
        if llm_config:
            api_key, _, _ = await self.get_byok_overrides_async(llm_config)
            if api_key:
                return Client(
                    api_key=api_key,
                    http_options=HttpOptions(timeout=timeout_ms),
                )
        return Client(
            vertexai=True,
            project=model_settings.google_cloud_project,
            location=model_settings.google_cloud_location,
            http_options=HttpOptions(api_version="v1", timeout=timeout_ms),
        )

    def _provider_prefix(self) -> str:
        return f"[{self.provider_label}]"

    def _provider_name(self) -> str:
        return self.provider_label

    @trace_method
    def request(self, request_data: dict, llm_config: LLMConfig) -> dict:
        """
        Performs underlying request to llm and returns raw response.
        """
        try:
            client = self._get_client(llm_config)
            response = client.models.generate_content(
                model=llm_config.model,
                contents=request_data["contents"],
                config=request_data["config"],
            )
            return response.model_dump()
        except pydantic_core._pydantic_core.ValidationError as e:
            # Handle Pydantic validation errors from the Google SDK
            # This occurs when tool schemas contain unsupported fields
            logger.error(
                f"Pydantic validation error when calling {self._provider_name()} API. Tool schema contains unsupported fields. Error: {e}"
            )
            raise LLMBadRequestError(
                message=f"Invalid tool schema for {self._provider_name()}: Tool parameters contain unsupported fields. "
                f"Common issues: 'const', 'default', 'additionalProperties' are not supported by Google AI. "
                f"Please check your tool definitions. Error: {str(e)}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
            )
        except Exception as e:
            raise self.handle_llm_error(e, llm_config=llm_config)

    @trace_method
    async def request_async(self, request_data: dict, llm_config: LLMConfig) -> dict:
        """
        Performs underlying request to llm and returns raw response.
        """
        request_data = sanitize_unicode_surrogates(request_data)

        client = await self._get_client_async(llm_config)

        # Gemini 2.5 models will often return MALFORMED_FUNCTION_CALL, force a retry
        # https://github.com/googleapis/python-aiplatform/issues/4472
        retry_count = 1
        should_retry = True
        response_data = None
        while should_retry and retry_count <= self.MAX_RETRIES:
            try:
                response = await client.aio.models.generate_content(
                    model=llm_config.model,
                    contents=request_data["contents"],
                    config=request_data["config"],
                )
            except pydantic_core._pydantic_core.ValidationError as e:
                # Handle Pydantic validation errors from the Google SDK
                # This occurs when tool schemas contain unsupported fields
                logger.error(
                    f"Pydantic validation error when calling {self._provider_name()} API. "
                    f"Tool schema contains unsupported fields. Error: {e}"
                )
                raise LLMBadRequestError(
                    message=f"Invalid tool schema for {self._provider_name()}: Tool parameters contain unsupported fields. "
                    f"Common issues: 'const', 'default', 'additionalProperties' are not supported by Google AI. "
                    f"Please check your tool definitions. Error: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                )
            except errors.APIError as e:
                # Retry on 503 and 500 errors as well, usually ephemeral from Gemini
                if e.code == 503 or e.code == 500 or e.code == 504:
                    logger.warning(f"Received {e}, retrying {retry_count}/{self.MAX_RETRIES}")
                    retry_count += 1
                    if retry_count > self.MAX_RETRIES:
                        raise self.handle_llm_error(e, llm_config=llm_config)
                    continue
                raise self.handle_llm_error(e, llm_config=llm_config)
            except Exception as e:
                raise self.handle_llm_error(e, llm_config=llm_config)
            response_data = response.model_dump()
            is_malformed_function_call = self.is_malformed_function_call(response_data)
            if is_malformed_function_call:
                logger.warning(
                    f"Received FinishReason.MALFORMED_FUNCTION_CALL in response for {llm_config.model}, retrying {retry_count}/{self.MAX_RETRIES}"
                )
                # Modify the last message if it's a heartbeat to include warning about special characters
                if request_data["contents"] and len(request_data["contents"]) > 0:
                    last_message = request_data["contents"][-1]
                    if last_message.get("role") == "user" and last_message.get("parts"):
                        for part in last_message["parts"]:
                            if "text" in part:
                                try:
                                    # Try to parse as JSON to check if it's a heartbeat
                                    message_json = json_loads(part["text"])
                                    if message_json.get("type") == "heartbeat" and "reason" in message_json:
                                        # Append warning to the reason
                                        warning = f" RETRY {retry_count}/{self.MAX_RETRIES} ***DO NOT USE SPECIAL CHARACTERS OR QUOTATIONS INSIDE FUNCTION CALL ARGUMENTS. IF YOU MUST, MAKE SURE TO ESCAPE THEM PROPERLY***"
                                        message_json["reason"] = message_json["reason"] + warning
                                        # Update the text with modified JSON
                                        part["text"] = json_dumps(message_json)
                                        logger.warning(
                                            f"Modified heartbeat message with special character warning for retry {retry_count}/{self.MAX_RETRIES}"
                                        )
                                except (json.JSONDecodeError, TypeError, AttributeError):
                                    # Not a JSON message or not a heartbeat, skip modification
                                    pass

            should_retry = is_malformed_function_call
            retry_count += 1

        if response_data is None:
            raise RuntimeError("Failed to get response data after all retries")
        return response_data

    @trace_method
    async def stream_async(self, request_data: dict, llm_config: LLMConfig) -> AsyncIterator[GenerateContentResponse]:
        request_data = sanitize_unicode_surrogates(request_data)

        client = await self._get_client_async(llm_config)

        try:
            response = await client.aio.models.generate_content_stream(
                model=llm_config.model,
                contents=request_data["contents"],
                config=request_data["config"],
            )
        except pydantic_core._pydantic_core.ValidationError as e:
            # Handle Pydantic validation errors from the Google SDK
            # This occurs when tool schemas contain unsupported fields
            logger.error(
                f"Pydantic validation error when calling {self._provider_name()} API. Tool schema contains unsupported fields. Error: {e}"
            )
            raise LLMBadRequestError(
                message=f"Invalid tool schema for {self._provider_name()}: Tool parameters contain unsupported fields. "
                f"Common issues: 'const', 'default', 'additionalProperties' are not supported by Google AI. "
                f"Please check your tool definitions. Error: {str(e)}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
            )
        except errors.APIError as e:
            raise self.handle_llm_error(e)
        except Exception as e:
            logger.error(f"Error streaming {self._provider_name()} request: {e} with request data: {json.dumps(request_data)}")
            raise e
        # Direct yield - keeps response alive in generator's local scope throughout iteration
        # This is required because the SDK's connection lifecycle is tied to the response object
        try:
            async for chunk in response:
                yield chunk
        except errors.ClientError as e:
            if e.code == 499:
                logger.info(f"{self._provider_prefix()} Stream cancelled by client (499): {e}")
                return
            raise self.handle_llm_error(e, llm_config=llm_config)
        except errors.APIError as e:
            raise self.handle_llm_error(e, llm_config=llm_config)

    @staticmethod
    def add_dummy_model_messages(messages: List[dict]) -> List[dict]:
        """Google AI API requires all function call returns are immediately followed by a 'model' role message.

        In Letta, the 'model' will often call a function (e.g. send_message) that itself yields to the user,
        so there is no natural follow-up 'model' role message.

        To satisfy the Google AI API restrictions, we can add a dummy 'yield' message
        with role == 'model' that is placed in-betweeen and function output
        (role == 'tool') and user message (role == 'user').
        """
        dummy_yield_message = {
            "role": "model",
            "parts": [{"text": f"{NON_USER_MSG_PREFIX}Function call returned, waiting for user response."}],
        }
        messages_with_padding = []
        for i, message in enumerate(messages):
            messages_with_padding.append(message)
            # Check if the current message role is 'tool' and the next message role is 'user'
            if message["role"] in ["tool", "function"] and (i + 1 < len(messages) and messages[i + 1]["role"] == "user"):
                messages_with_padding.append(dummy_yield_message)

        return messages_with_padding

    def _clean_google_ai_schema_properties(self, schema_part: dict):
        """Recursively clean schema parts to remove unsupported Google AI keywords."""
        if not isinstance(schema_part, dict):
            return

        # Per https://ai.google.dev/gemini-api/docs/function-calling?example=meeting#notes_and_limitations
        # * Only a subset of the OpenAPI schema is supported.
        # * Supported parameter types in Python are limited.
        unsupported_keys = ["default", "exclusiveMaximum", "exclusiveMinimum", "additionalProperties", "$schema", "const", "$ref"]
        keys_to_remove_at_this_level = [key for key in unsupported_keys if key in schema_part]
        for key_to_remove in keys_to_remove_at_this_level:
            logger.debug(f"Removing unsupported keyword 	'{key_to_remove}' from schema part.")
            del schema_part[key_to_remove]

        if schema_part.get("type") == "string" and "format" in schema_part:
            allowed_formats = ["enum", "date-time"]
            if schema_part["format"] not in allowed_formats:
                logger.warning(f"Removing unsupported format 	'{schema_part['format']}' for string type. Allowed: {allowed_formats}")
                del schema_part["format"]

        # Check properties within the current level
        if "properties" in schema_part and isinstance(schema_part["properties"], dict):
            for prop_name, prop_schema in schema_part["properties"].items():
                self._clean_google_ai_schema_properties(prop_schema)

        # Check items within arrays
        if "items" in schema_part and isinstance(schema_part["items"], dict):
            self._clean_google_ai_schema_properties(schema_part["items"])

        # Check within anyOf, allOf, oneOf lists
        for key in ["anyOf", "allOf", "oneOf"]:
            if key in schema_part and isinstance(schema_part[key], list):
                for item_schema in schema_part[key]:
                    self._clean_google_ai_schema_properties(item_schema)

    def _resolve_json_schema_refs(self, schema: dict, defs: dict | None = None) -> dict:
        """
        Recursively resolve $ref in JSON schema by inlining definitions.
        Google GenAI SDK does not support $ref.
        """
        if defs is None:
            # Look for definitions at the top level
            defs = schema.get("$defs") or schema.get("definitions") or {}

        if not isinstance(schema, dict):
            return schema

        # If this is a ref, resolve it
        if "$ref" in schema:
            ref = schema["$ref"]
            if isinstance(ref, str):
                for prefix in ("#/$defs/", "#/definitions/"):
                    if ref.startswith(prefix):
                        ref_name = ref.split("/")[-1]
                        if ref_name in defs:
                            resolved = defs[ref_name].copy()
                            return self._resolve_json_schema_refs(resolved, defs)
                        break

            logger.warning(f"Could not resolve $ref '{ref}' in schema — will be stripped by schema cleaner")

        # Recursively process children
        new_schema = schema.copy()

        # We need to remove $defs/definitions from the output schema as Google doesn't support them
        if "$defs" in new_schema:
            del new_schema["$defs"]
        if "definitions" in new_schema:
            del new_schema["definitions"]

        for k, v in new_schema.items():
            if isinstance(v, dict):
                new_schema[k] = self._resolve_json_schema_refs(v, defs)
            elif isinstance(v, list):
                new_schema[k] = [self._resolve_json_schema_refs(i, defs) if isinstance(i, dict) else i for i in v]

        return new_schema

    def convert_tools_to_google_ai_format(self, tools: List[Tool], llm_config: LLMConfig) -> List[dict]:
        """
        OpenAI style:
        "tools": [{
            "type": "function",
            "function": {
                "name": "find_movies",
                "description": "find ....",
                "parameters": {
                "type": "object",
                "properties": {
                    PARAM: {
                    "type": PARAM_TYPE,  # eg "string"
                    "description": PARAM_DESCRIPTION,
                    },
                    ...
                },
                "required": List[str],
                }
            }
        }
        ]

        Google AI style:
        "tools": [{
            "functionDeclarations": [{
            "name": "find_movies",
            "description": "find movie titles currently playing in theaters based on any description, genre, title words, etc.",
            "parameters": {
                "type": "OBJECT",
                "properties": {
                "location": {
                    "type": "STRING",
                    "description": "The city and state, e.g. San Francisco, CA or a zip code e.g. 95616"
                },
                "description": {
                    "type": "STRING",
                    "description": "Any kind of description including category or genre, title words, attributes, etc."
                }
                },
                "required": ["description"]
            }
            }, {
            "name": "find_theaters",
            ...
        """
        function_list = [
            dict(
                name=t.function.name,
                description=t.function.description,
                # Deep copy parameters to avoid modifying the original Tool object
                parameters=copy.deepcopy(t.function.parameters) if t.function.parameters else {},
            )
            for t in tools
        ]

        # Add inner thoughts if needed
        for func in function_list:
            # Note: Google AI API used to have weird casing requirements, but not any more

            # Google AI API only supports a subset of OpenAPI 3.0, so unsupported params must be cleaned
            if "parameters" in func and isinstance(func["parameters"], dict):
                # Resolve $ref in schema because Google AI SDK doesn't support them
                func["parameters"] = self._resolve_json_schema_refs(func["parameters"])
                self._clean_google_ai_schema_properties(func["parameters"])

            # Add inner thoughts
            if llm_config.put_inner_thoughts_in_kwargs:
                from letta.local_llm.constants import INNER_THOUGHTS_KWARG_DESCRIPTION, INNER_THOUGHTS_KWARG_VERTEX

                func["parameters"]["properties"][INNER_THOUGHTS_KWARG_VERTEX] = {
                    "type": "string",
                    "description": INNER_THOUGHTS_KWARG_DESCRIPTION,
                }
                func["parameters"]["required"].append(INNER_THOUGHTS_KWARG_VERTEX)

        return [{"functionDeclarations": function_list}]

    @trace_method
    def build_request_data(
        self,
        agent_type: AgentType,  # if react, use native content + strip heartbeats
        messages: List[PydanticMessage],
        llm_config: LLMConfig,
        tools: List[dict],
        force_tool_call: Optional[str] = None,
        requires_subsequent_tool_call: bool = False,
        tool_return_truncation_chars: Optional[int] = None,
    ) -> dict:
        """
        Constructs a request object in the expected data format for this client.
        """
        # NOTE: forcing inner thoughts in kwargs off
        if agent_type == AgentType.letta_v1_agent:
            llm_config.put_inner_thoughts_in_kwargs = False

        if tools:
            tool_objs = [Tool(type="function", function=t) for t in tools]
            tool_names = [t.function.name for t in tool_objs]
            # Convert to the exact payload style Google expects
            formatted_tools = self.convert_tools_to_google_ai_format(tool_objs, llm_config)
        else:
            formatted_tools = []
            tool_names = []

        contents = self.add_dummy_model_messages(
            PydanticMessage.to_google_dicts_from_list(
                messages,
                current_model=llm_config.model,
                put_inner_thoughts_in_kwargs=False if agent_type == AgentType.letta_v1_agent else True,
                native_content=True if agent_type == AgentType.letta_v1_agent else False,
            ),
        )

        request_data = {
            "contents": contents,
            "config": {
                "temperature": llm_config.temperature,
                "tools": formatted_tools,
            },
        }
        # Make tokens is optional
        if llm_config.max_tokens:
            request_data["config"]["max_output_tokens"] = llm_config.max_tokens

        if len(tool_names) == 1 and settings.use_vertex_structured_outputs_experimental:
            request_data["config"]["response_mime_type"] = "application/json"
            request_data["config"]["response_schema"] = self.get_function_call_response_schema(tools[0])
            del request_data["config"]["tools"]
        elif tools:
            if agent_type == AgentType.letta_v1_agent:
                # don't require tools
                tool_call_mode = FunctionCallingConfigMode.AUTO
                tool_config = ToolConfig(
                    function_calling_config=FunctionCallingConfig(
                        mode=tool_call_mode,
                    )
                )
            else:
                # require tools
                tool_call_mode = FunctionCallingConfigMode.ANY
                tool_config = ToolConfig(
                    function_calling_config=FunctionCallingConfig(
                        mode=tool_call_mode,
                        # Provide the list of tools (though empty should also work, it seems not to)
                        allowed_function_names=tool_names,
                    )
                )

            request_data["config"]["tool_config"] = tool_config.model_dump()

        # https://ai.google.dev/gemini-api/docs/thinking#set-budget
        # 2.5 Pro
        #   - Default: dynamic thinking
        #   - Dynamic thinking that cannot be disabled
        #   - Range: -1 (for dynamic), or 128-32768
        # 2.5 Flash
        #   - Default: dynamic thinking
        #   - Dynamic thinking that *can* be disabled
        #   - Range: -1, 0, or 0-24576
        # 2.5 Flash Lite
        #   - Default: no thinking
        #   - Dynamic thinking that *can* be disabled
        #   - Range: -1, 0, or 512-24576
        # TODO when using v3 agent loop, properly support the native thinking in Gemini

        # Add thinking_config for all Gemini reasoning models (2.5 series)
        # If enable_reasoner is False, set thinking_budget to 0
        # Otherwise, use the value from max_reasoning_tokens
        if self.is_reasoning_model(llm_config) or "flash" in llm_config.model:
            if llm_config.model.startswith("gemini-3"):
                # letting thinking_level to default to high by not specifying thinking_budget
                thinking_config = ThinkingConfig(include_thoughts=True)
            else:
                # Gemini reasoning models may fail to call tools even with FunctionCallingConfigMode.ANY if thinking is fully disabled, set to minimum to prevent tool call failure
                thinking_budget = (
                    llm_config.max_reasoning_tokens if llm_config.enable_reasoner else self.get_thinking_budget(llm_config.model)
                )
                if thinking_budget <= 0:
                    logger.warning(
                        f"Thinking budget of {thinking_budget} for Gemini reasoning model {llm_config.model}, this will likely cause tool call failures"
                    )
                    # For models that require thinking mode (2.5 Pro, 3.x), override with minimum valid budget
                    if llm_config.model.startswith("gemini-2.5-pro"):
                        thinking_budget = 128
                        logger.warning(
                            f"Overriding thinking_budget to {thinking_budget} for model {llm_config.model} which requires thinking mode"
                        )
                thinking_config = ThinkingConfig(
                    thinking_budget=(thinking_budget),
                    include_thoughts=(thinking_budget > 1),
                )
            request_data["config"]["thinking_config"] = thinking_config.model_dump()

        return request_data

    def extract_usage_statistics(self, response_data: dict | None, llm_config: LLMConfig) -> LettaUsageStatistics:
        """Extract usage statistics from Gemini response and return as LettaUsageStatistics."""
        if not response_data:
            return LettaUsageStatistics()

        response = GenerateContentResponse(**response_data)
        if not response.usage_metadata:
            return LettaUsageStatistics()

        cached_tokens = None
        if (
            hasattr(response.usage_metadata, "cached_content_token_count")
            and response.usage_metadata.cached_content_token_count is not None
        ):
            cached_tokens = response.usage_metadata.cached_content_token_count

        reasoning_tokens = None
        if hasattr(response.usage_metadata, "thoughts_token_count") and response.usage_metadata.thoughts_token_count is not None:
            reasoning_tokens = response.usage_metadata.thoughts_token_count

        return LettaUsageStatistics(
            prompt_tokens=response.usage_metadata.prompt_token_count or 0,
            completion_tokens=response.usage_metadata.candidates_token_count or 0,
            total_tokens=response.usage_metadata.total_token_count or 0,
            cached_input_tokens=cached_tokens,
            reasoning_tokens=reasoning_tokens,
        )

    @trace_method
    async def convert_response_to_chat_completion(
        self,
        response_data: dict,
        input_messages: List[PydanticMessage],
        llm_config: LLMConfig,
    ) -> ChatCompletionResponse:
        """
        Converts custom response format from llm client into an OpenAI
        ChatCompletionsResponse object.

        Example:
        {
        "candidates": [
            {
                "content": {
                    "parts": [
                        {
                            "text": " OK. Barbie is showing in two theaters in Mountain View, CA: AMC Mountain View 16 and Regal Edwards 14."
                        }
                    ]
                }
            }
        ],
        "usageMetadata": {
            "promptTokenCount": 9,
            "candidatesTokenCount": 27,
            "totalTokenCount": 36
        }
        }
        """
        response = GenerateContentResponse(**response_data)
        try:
            choices = []
            index = 0
            for candidate in response.candidates:
                content = candidate.content

                if content is None or content.role is None or content.parts is None:
                    # This means the response is malformed like MALFORMED_FUNCTION_CALL
                    if candidate.finish_reason == "MALFORMED_FUNCTION_CALL":
                        raise LLMServerError(f"Malformed response from {self._provider_name()}: {candidate.finish_reason}")
                    else:
                        raise LLMServerError(f"Invalid response data from {self._provider_name()}: {candidate.model_dump()}")

                role = content.role
                assert role == "model", f"Unknown role in response: {role}"

                parts = content.parts

                # NOTE: we aren't properly supported multi-parts here anyways (we're just appending choices),
                #       so let's disable it for now

                # NOTE(Apr 9, 2025): there's a very strange bug on 2.5 where the response has a part with broken text
                # {'candidates': [{'content': {'parts': [{'functionCall': {'name': 'send_message', 'args': {'request_heartbeat': False, 'message': 'Hello! How can I make your day better?', 'inner_thoughts': 'User has initiated contact. Sending a greeting.'}}}], 'role': 'model'}, 'finishReason': 'STOP', 'avgLogprobs': -0.25891534213362066}], 'usageMetadata': {'promptTokenCount': 2493, 'candidatesTokenCount': 29, 'totalTokenCount': 2522, 'promptTokensDetails': [{'modality': 'TEXT', 'tokenCount': 2493}], 'candidatesTokensDetails': [{'modality': 'TEXT', 'tokenCount': 29}]}, 'modelVersion': 'gemini-1.5-pro-002'}
                # To patch this, if we have multiple parts we can take the last one
                # Unless parallel tool calling is enabled, in which case multiple parts may be intentional
                if len(parts) > 1 and not llm_config.enable_reasoner and not llm_config.parallel_tool_calls:
                    logger.warning(f"Unexpected multiple parts in response from Google AI: {parts}")
                    # only truncate if reasoning is off and parallel tool calls are disabled
                    parts = [parts[-1]]

                # TODO support parts / multimodal
                # Parallel tool calling is now supported when llm_config.parallel_tool_calls is enabled
                openai_response_message = None
                for response_message in parts:
                    # Convert the actual message style to OpenAI style
                    if response_message.function_call:
                        function_call = response_message.function_call
                        function_name = function_call.name
                        function_args = function_call.args
                        assert isinstance(function_args, dict), function_args

                        # TODO this is kind of funky - really, we should be passing 'native_content' as a kwarg to fork behavior
                        inner_thoughts = response_message.text
                        if llm_config.put_inner_thoughts_in_kwargs:
                            # NOTE: this also involves stripping the inner monologue out of the function
                            from letta.local_llm.constants import INNER_THOUGHTS_KWARG_VERTEX

                            assert INNER_THOUGHTS_KWARG_VERTEX in function_args, (
                                f"Couldn't find inner thoughts in function args:\n{function_call}"
                            )
                            inner_thoughts = function_args.pop(INNER_THOUGHTS_KWARG_VERTEX)
                            assert inner_thoughts is not None, f"Expected non-null inner thoughts function arg:\n{function_call}"
                        else:
                            pass
                            # inner_thoughts = None
                            # inner_thoughts = response_message.text

                        # Google AI API doesn't generate tool call IDs
                        tool_call = ToolCall(
                            id=get_tool_call_id(),
                            type="function",
                            function=FunctionCall(
                                name=function_name,
                                arguments=clean_json_string_extra_backslash(json_dumps(function_args)),
                            ),
                        )

                        if openai_response_message is None:
                            openai_response_message = Message(
                                role="assistant",  # NOTE: "model" -> "assistant"
                                content=inner_thoughts,
                                tool_calls=[tool_call],
                            )
                            if response_message.thought_signature:
                                thought_signature = base64.b64encode(response_message.thought_signature).decode("utf-8")
                                openai_response_message.reasoning_content_signature = thought_signature
                        else:
                            openai_response_message.content = inner_thoughts
                            if openai_response_message.tool_calls is None:
                                openai_response_message.tool_calls = []
                            openai_response_message.tool_calls.append(tool_call)
                            if response_message.thought_signature:
                                thought_signature = base64.b64encode(response_message.thought_signature).decode("utf-8")
                                openai_response_message.reasoning_content_signature = thought_signature

                    else:
                        if response_message.thought:
                            if openai_response_message is None:
                                openai_response_message = Message(
                                    role="assistant",  # NOTE: "model" -> "assistant"
                                    reasoning_content=response_message.text,
                                )
                            else:
                                openai_response_message.reasoning_content = response_message.text
                        try:
                            # Structured output tool call
                            function_call = json_loads(response_message.text)

                            # Access dict keys - will raise TypeError/KeyError if not a dict or missing keys
                            function_name = function_call["name"]
                            function_args = function_call["args"]
                            assert isinstance(function_args, dict), function_args

                            # NOTE: this also involves stripping the inner monologue out of the function
                            if llm_config.put_inner_thoughts_in_kwargs:
                                from letta.local_llm.constants import INNER_THOUGHTS_KWARG_VERTEX

                                assert INNER_THOUGHTS_KWARG_VERTEX in function_args, (
                                    f"Couldn't find inner thoughts in function args:\n{function_call}"
                                )
                                inner_thoughts = function_args.pop(INNER_THOUGHTS_KWARG_VERTEX)
                                assert inner_thoughts is not None, f"Expected non-null inner thoughts function arg:\n{function_call}"
                            else:
                                inner_thoughts = None

                            # Google AI API doesn't generate tool call IDs
                            tool_call = ToolCall(
                                id=get_tool_call_id(),
                                type="function",
                                function=FunctionCall(
                                    name=function_name,
                                    arguments=clean_json_string_extra_backslash(json_dumps(function_args)),
                                ),
                            )
                            if openai_response_message is None:
                                openai_response_message = Message(
                                    role="assistant",  # NOTE: "model" -> "assistant"
                                    content=inner_thoughts,
                                    tool_calls=[tool_call],
                                )
                            else:
                                openai_response_message.content = inner_thoughts
                                if openai_response_message.tool_calls is None:
                                    openai_response_message.tool_calls = []
                                openai_response_message.tool_calls.append(tool_call)

                        except (json.decoder.JSONDecodeError, ValueError, TypeError, KeyError, AssertionError) as e:
                            if candidate.finish_reason == "MAX_TOKENS":
                                raise LLMServerError("Could not parse response data from LLM: exceeded max token limit")
                            # Log the parsing error for debugging
                            logger.warning(
                                f"Failed to parse structured output from LLM response: {e}. Response text: {response_message.text[:500]}"
                            )
                            # Inner thoughts are the content by default
                            inner_thoughts = response_message.text

                            # Google AI API doesn't generate tool call IDs
                            if openai_response_message is None:
                                openai_response_message = Message(
                                    role="assistant",  # NOTE: "model" -> "assistant"
                                    content=inner_thoughts,
                                )
                            else:
                                openai_response_message.content = inner_thoughts
                            if response_message.thought_signature:
                                thought_signature = base64.b64encode(response_message.thought_signature).decode("utf-8")
                                openai_response_message.reasoning_content_signature = thought_signature

                    # Google AI API uses different finish reason strings than OpenAI
                    # OpenAI: 'stop', 'length', 'function_call', 'content_filter', null
                    #   see: https://platform.openai.com/docs/guides/text-generation/chat-completions-api
                    # Google AI API: FINISH_REASON_UNSPECIFIED, STOP, MAX_TOKENS, SAFETY, RECITATION, OTHER
                    #   see: https://ai.google.dev/api/python/google/ai/generativelanguage/Candidate/FinishReason
                    finish_reason = candidate.finish_reason.value
                    if finish_reason == "STOP":
                        openai_finish_reason = (
                            "function_call"
                            if openai_response_message.tool_calls is not None and len(openai_response_message.tool_calls) > 0
                            else "stop"
                        )
                    elif finish_reason == "MAX_TOKENS":
                        openai_finish_reason = "length"
                    elif finish_reason == "SAFETY":
                        openai_finish_reason = "content_filter"
                    elif finish_reason == "RECITATION":
                        openai_finish_reason = "content_filter"
                    else:
                        raise LLMServerError(f"Unrecognized finish reason in Google AI response: {finish_reason}")

                    choices.append(
                        Choice(
                            finish_reason=openai_finish_reason,
                            index=index,
                            message=openai_response_message,
                        )
                    )
                    index += 1

            # if len(choices) > 1:
            #     raise UserWarning(f"Unexpected number of candidates in response (expected 1, got {len(choices)})")

            # NOTE: some of the Google AI APIs show UsageMetadata in the response, but it seems to not exist?
            #  "usageMetadata": {
            #     "promptTokenCount": 9,
            #     "candidatesTokenCount": 27,
            #     "totalTokenCount": 36
            #   }
            usage = None
            if response.usage_metadata:
                # Extract usage via centralized method
                from letta.schemas.enums import ProviderType

                usage = self.extract_usage_statistics(response_data, llm_config).to_usage(ProviderType.google_ai)
            else:
                # Count it ourselves using the Gemini token counting API
                assert input_messages is not None, "Didn't get UsageMetadata from the API response, so input_messages is required"
                google_messages = PydanticMessage.to_google_dicts_from_list(input_messages, current_model=llm_config.model)
                prompt_tokens = await self.count_tokens(messages=google_messages, model=llm_config.model)
                # For completion tokens, wrap the response content in Google format
                completion_content = [{"role": "model", "parts": [{"text": json_dumps(openai_response_message.model_dump())}]}]
                completion_tokens = await self.count_tokens(messages=completion_content, model=llm_config.model)
                total_tokens = prompt_tokens + completion_tokens
                usage = UsageStatistics(
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                    total_tokens=total_tokens,
                )

            response_id = str(uuid.uuid4())
            return ChatCompletionResponse(
                id=response_id,
                choices=choices,
                model=llm_config.model,  # NOTE: Google API doesn't pass back model in the response
                created=get_utc_time_int(),
                usage=usage,
            )
        except KeyError as e:
            raise e

    def get_function_call_response_schema(self, tool: dict) -> dict:
        return {
            "type": "OBJECT",
            "properties": {
                "name": {"type": "STRING", "enum": [tool["name"]]},
                "args": {
                    "type": "OBJECT",
                    "properties": tool["parameters"]["properties"],
                    "required": tool["parameters"]["required"],
                },
            },
            "propertyOrdering": ["name", "args"],
            "required": ["name", "args"],
        }

    # https://ai.google.dev/gemini-api/docs/thinking#set-budget
    # | Model           | Default setting                                                   | Range        | Disable thinking           | Turn on dynamic thinking|
    # |-----------------|-------------------------------------------------------------------|--------------|----------------------------|-------------------------|
    # | 2.5 Pro         | Dynamic thinking: Model decides when and how much to think        | 128-32768    | N/A: Cannot disable        | thinkingBudget = -1     |
    # | 2.5 Flash       | Dynamic thinking: Model decides when and how much to think        | 0-24576      | thinkingBudget = 0         | thinkingBudget = -1     |
    # | 2.5 Flash Lite  | Model does not think                                              | 512-24576    | thinkingBudget = 0         | thinkingBudget = -1     |
    # | 3.x             | Dynamic thinking: Model decides when and how much to think        | 128-?        | N/A: Cannot disable        | thinkingBudget = -1     |
    def get_thinking_budget(self, model: str) -> bool:
        if model_settings.gemini_force_minimum_thinking_budget:
            if all(substring in model for substring in ["2.5", "flash", "lite"]):
                return 512
            elif all(substring in model for substring in ["2.5", "flash"]):
                return 1
        # Gemini 3 and 2.5 Pro require thinking mode and cannot have budget 0
        if model.startswith("gemini-3") or model.startswith("gemini-2.5-pro"):
            return 128  # Minimum valid budget for models that require thinking
        return 0

    def is_reasoning_model(self, llm_config: LLMConfig) -> bool:
        return (
            llm_config.model.startswith("gemini-2.5-flash")
            or llm_config.model.startswith("gemini-2.5-pro")
            or llm_config.model.startswith("gemini-3")
        )

    def is_malformed_function_call(self, response_data: dict) -> dict:
        response = GenerateContentResponse(**response_data)
        for candidate in response.candidates:
            content = candidate.content
            if content is None or content.role is None or content.parts is None:
                return candidate.finish_reason == "MALFORMED_FUNCTION_CALL"
        return False

    @trace_method
    def handle_llm_error(self, e: Exception, llm_config: Optional[LLMConfig] = None) -> Exception:
        is_byok = (llm_config.provider_category == ProviderCategory.byok) if llm_config else None

        # Handle Google GenAI specific errors
        if isinstance(e, errors.ClientError):
            if e.code == 499:
                logger.info(f"{self._provider_prefix()} Request cancelled by client (499): {e}")
                return LLMConnectionError(
                    message=f"Request to {self._provider_name()} was cancelled (client disconnected): {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={"status_code": 499, "cause": "client_cancelled", "is_byok": is_byok},
                )

            logger.warning(f"{self._provider_prefix()} Client error ({e.code}): {e}")

            # Handle specific error codes
            if e.code == 400:
                error_str = str(e).lower()
                if ("context" in error_str or "token count" in error_str or "tokens allowed" in error_str) and (
                    "exceed" in error_str or "limit" in error_str or "too long" in error_str
                ):
                    return ContextWindowExceededError(
                        message=f"Bad request to {self._provider_name()} (context window exceeded): {str(e)}",
                        details={"is_byok": is_byok},
                    )
                else:
                    return LLMBadRequestError(
                        message=f"Bad request to {self._provider_name()}: {str(e)}",
                        code=ErrorCode.INVALID_ARGUMENT,
                        details={"is_byok": is_byok},
                    )
            elif e.code == 401:
                return LLMAuthenticationError(
                    message=f"Authentication failed with {self._provider_name()}: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={"is_byok": is_byok},
                )
            elif e.code == 403:
                return LLMPermissionDeniedError(
                    message=f"Permission denied by {self._provider_name()}: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={"is_byok": is_byok},
                )
            elif e.code == 404:
                return LLMNotFoundError(
                    message=f"Resource not found in {self._provider_name()}: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={"is_byok": is_byok},
                )
            elif e.code == 408:
                return LLMTimeoutError(
                    message=f"Request to {self._provider_name()} timed out: {str(e)}",
                    code=ErrorCode.TIMEOUT,
                    details={"cause": str(e.__cause__) if e.__cause__ else None, "is_byok": is_byok},
                )
            elif e.code == 402 or is_insufficient_credits_message(str(e)):
                msg = str(e)
                return LLMInsufficientCreditsError(
                    message=f"Insufficient credits (BYOK): {msg}" if is_byok else f"Insufficient credits: {msg}",
                    code=ErrorCode.PAYMENT_REQUIRED,
                    details={"status_code": e.code, "is_byok": is_byok},
                )
            elif e.code == 422:
                return LLMUnprocessableEntityError(
                    message=f"Invalid request content for {self._provider_name()}: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={"is_byok": is_byok},
                )
            elif e.code == 429:
                logger.warning(f"{self._provider_prefix()} Rate limited (429). Consider backoff.")
                return LLMRateLimitError(
                    message=f"Rate limited by {self._provider_name()}: {str(e)}",
                    code=ErrorCode.RATE_LIMIT_EXCEEDED,
                    details={"is_byok": is_byok},
                )
            else:
                return LLMServerError(
                    message=f"{self._provider_name()} client error: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={
                        "status_code": e.code,
                        "response_json": getattr(e, "response_json", None),
                        "is_byok": is_byok,
                    },
                )

        if isinstance(e, errors.ServerError):
            logger.warning(f"{self._provider_prefix()} Server error ({e.code}): {e}")

            # Handle specific server error codes
            if e.code == 500:
                return LLMServerError(
                    message=f"{self._provider_name()} internal server error: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={
                        "status_code": e.code,
                        "response_json": getattr(e, "response_json", None),
                        "is_byok": is_byok,
                    },
                )
            elif e.code == 502:
                return LLMConnectionError(
                    message=f"Bad gateway from {self._provider_name()}: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={"cause": str(e.__cause__) if e.__cause__ else None, "is_byok": is_byok},
                )
            elif e.code == 503:
                return LLMServerError(
                    message=f"{self._provider_name()} service unavailable: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={
                        "status_code": e.code,
                        "response_json": getattr(e, "response_json", None),
                        "is_byok": is_byok,
                    },
                )
            elif e.code == 504:
                return LLMTimeoutError(
                    message=f"Gateway timeout from {self._provider_name()}: {str(e)}",
                    code=ErrorCode.TIMEOUT,
                    details={"cause": str(e.__cause__) if e.__cause__ else None, "is_byok": is_byok},
                )
            else:
                return LLMServerError(
                    message=f"{self._provider_name()} server error: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={
                        "status_code": e.code,
                        "response_json": getattr(e, "response_json", None),
                        "is_byok": is_byok,
                    },
                )

        if isinstance(e, errors.APIError):
            logger.warning(f"{self._provider_prefix()} API error ({e.code}): {e}")
            return LLMServerError(
                message=f"{self._provider_name()} API error: {str(e)}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                details={
                    "status_code": e.code,
                    "response_json": getattr(e, "response_json", None),
                    "is_byok": is_byok,
                },
            )

        # Handle httpx.RemoteProtocolError which can occur during streaming
        # when the remote server closes the connection unexpectedly
        # (e.g., "peer closed connection without sending complete message body")
        if isinstance(e, httpx.RemoteProtocolError):
            logger.warning(f"{self._provider_prefix()} Remote protocol error during streaming: {e}")
            return LLMConnectionError(
                message=f"Connection error during {self._provider_name()} streaming: {str(e)}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                details={"cause": str(e.__cause__) if e.__cause__ else None, "is_byok": is_byok},
            )

        # Handle httpx network errors which can occur during streaming
        # when the connection is unexpectedly closed while reading/writing
        if isinstance(e, (httpx.ReadError, httpx.WriteError, httpx.ConnectError)):
            logger.warning(f"{self._provider_prefix()} Network error during streaming: {type(e).__name__}: {e}")
            return LLMConnectionError(
                message=f"Network error during {self._provider_name()} streaming: {str(e)}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                details={"cause": str(e.__cause__) if e.__cause__ else None, "error_type": type(e).__name__, "is_byok": is_byok},
            )

        # Handle connection-related errors
        if "connection" in str(e).lower() or "timeout" in str(e).lower():
            logger.warning(f"{self._provider_prefix()} Connection/timeout error: {e}")
            return LLMConnectionError(
                message=f"Failed to connect to {self._provider_name()}: {str(e)}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                details={"cause": str(e.__cause__) if e.__cause__ else None, "is_byok": is_byok},
            )

        # Fallback to base implementation for other errors
        return super().handle_llm_error(e, llm_config=llm_config)

    async def count_tokens(
        self, messages: List[dict] | None = None, model: str | None = None, tools: List[OpenAITool] | None = None
    ) -> int:
        """
        Count tokens for the given messages and tools using the Gemini token counting API.

        Args:
            messages: List of message dicts in Google AI format (with 'role' and 'parts' keys)
            model: The model to use for token counting (defaults to gemini-2.0-flash-lite)
            tools: List of OpenAI-style Tool objects to include in the count

        Returns:
            The total token count for the input
        """
        from letta.llm_api.google_constants import GOOGLE_MODEL_FOR_API_KEY_CHECK

        client = self._get_client()

        # Default model for token counting if not specified
        count_model = model or GOOGLE_MODEL_FOR_API_KEY_CHECK

        # Build the contents parameter
        # If no messages provided, use empty string (like the API key check)
        if messages is None or len(messages) == 0:
            contents = ""
        else:
            # Messages should already be in Google format (role + parts)
            contents = messages

        try:
            # Count message tokens
            result = await client.aio.models.count_tokens(
                model=count_model,
                contents=contents,
            )
            total_tokens = result.total_tokens

            # Count tool tokens separately by serializing to text
            # The Gemini count_tokens API doesn't support a tools parameter directly
            if tools and len(tools) > 0:
                # Serialize tools to JSON text and count those tokens
                tools_text = json.dumps([t.model_dump() for t in tools])
                tools_result = await client.aio.models.count_tokens(
                    model=count_model,
                    contents=tools_text,
                )
                total_tokens += tools_result.total_tokens

        except Exception as e:
            raise self.handle_llm_error(e)

        return total_tokens



from typing import Literal

from letta.log import get_logger

logger = get_logger(__name__)

from pydantic import Field

from letta.schemas.enums import ProviderCategory, ProviderType
from letta.schemas.llm_config import LLMConfig
from letta.schemas.providers.openai import OpenAIProvider

# Z.ai model context windows
# Reference: https://docs.z.ai/
# GLM-5 max context window is 200K tokens but max_output_tokens (default 16k) counts against that --> 180k
MODEL_CONTEXT_WINDOWS = {
    "glm-4.5": 128000,
    "glm-4.6": 180000,
    "glm-4.7": 180000,
    "glm-5": 180000,
    "glm-5-code": 180000,
}


class ZAIProvider(OpenAIProvider):
    """Z.ai (ZhipuAI) provider - https://docs.z.ai/"""

    provider_type: Literal[ProviderType.zai] = Field(ProviderType.zai, description="The type of the provider.")
    provider_category: ProviderCategory = Field(ProviderCategory.base, description="The category of the provider (base or byok)")
    api_key: str | None = Field(None, description="API key for the Z.ai API.", deprecated=True)
    base_url: str = Field("https://api.z.ai/api/paas/v4/", description="Base URL for the Z.ai API.")

    def get_model_context_window_size(self, model_name: str) -> int | None:
        # Z.ai doesn't return context window in the model listing,
        # this is hardcoded from documentation
        return MODEL_CONTEXT_WINDOWS.get(model_name)

    async def list_llm_models_async(self) -> list[LLMConfig]:
        from letta.llm_api.openai import openai_get_model_list_async

        api_key = await self.api_key_enc.get_plaintext_async() if self.api_key_enc else None
        response = await openai_get_model_list_async(self.base_url, api_key=api_key)

        data = response.get("data", response)

        configs = []
        for model in data:
            assert "id" in model, f"Z.ai model missing 'id' field: {model}"
            model_name = model["id"]

            # In case Z.ai starts supporting it in the future:
            if "context_length" in model:
                context_window_size = model["context_length"]
            else:
                context_window_size = self.get_model_context_window_size(model_name)

            if not context_window_size:
                logger.warning(f"Couldn't find context window size for model {model_name}")
                continue

            configs.append(
                LLMConfig(
                    model=model_name,
                    model_endpoint_type=self.provider_type.value,
                    model_endpoint=self.base_url,
                    context_window=context_window_size,
                    handle=self.get_handle(model_name),
                    max_tokens=self.get_default_max_output_tokens(model_name),
                    provider_name=self.name,
                    provider_category=self.provider_category,
                )
            )

        return configs



"""ChatGPT OAuth Client - handles requests to chatgpt.com/backend-api/codex/responses."""

import asyncio
import json
from typing import Any, AsyncIterator, Dict, List, Optional

import httpx
from openai.types.responses import (
    Response,
    ResponseCompletedEvent,
    ResponseContentPartAddedEvent,
    ResponseContentPartDoneEvent,
    ResponseCreatedEvent,
    ResponseFunctionCallArgumentsDeltaEvent,
    ResponseFunctionCallArgumentsDoneEvent,
    ResponseFunctionToolCall,
    ResponseInProgressEvent,
    ResponseOutputItemAddedEvent,
    ResponseOutputItemDoneEvent,
    ResponseOutputMessage,
    ResponseOutputText,
    ResponseReasoningItem,
    ResponseReasoningSummaryPartAddedEvent,
    ResponseReasoningSummaryPartDoneEvent,
    ResponseReasoningSummaryTextDeltaEvent,
    ResponseReasoningSummaryTextDoneEvent,
    ResponseTextDeltaEvent,
    ResponseTextDoneEvent,
)
from openai.types.responses.response_stream_event import ResponseStreamEvent

from letta.errors import (
    ContextWindowExceededError,
    ErrorCode,
    LettaError,
    LLMAuthenticationError,
    LLMBadRequestError,
    LLMConnectionError,
    LLMRateLimitError,
    LLMServerError,
    LLMTimeoutError,
)
from letta.helpers.json_helpers import sanitize_unicode_surrogates
from letta.llm_api.llm_client_base import LLMClientBase
from letta.log import get_logger
from letta.otel.tracing import trace_method
from letta.schemas.enums import AgentType, ProviderCategory
from letta.schemas.llm_config import LLMConfig
from letta.schemas.message import Message as PydanticMessage
from letta.schemas.openai.chat_completion_response import (
    ChatCompletionResponse,
)
from letta.schemas.providers.chatgpt_oauth import ChatGPTOAuthCredentials, ChatGPTOAuthProvider
from letta.schemas.usage import LettaUsageStatistics

logger = get_logger(__name__)

# ChatGPT Backend API endpoint
CHATGPT_CODEX_ENDPOINT = "https://chatgpt.com/backend-api/codex/responses"


class AsyncStreamWrapper:
    """Wraps an async generator to provide async context manager protocol.

    The OpenAI SDK's AsyncStream implements __aenter__ and __aexit__,
    but our custom SSE handler returns a raw async generator. This wrapper
    provides the context manager protocol so it can be used with 'async with'.
    """

    def __init__(self, generator: AsyncIterator[ResponseStreamEvent]):
        self._generator = generator

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # Close the generator if it has an aclose method
        if hasattr(self._generator, "aclose"):
            await self._generator.aclose()
        return False

    def __aiter__(self):
        return self

    async def __anext__(self) -> ResponseStreamEvent:
        return await self._generator.__anext__()


class ChatGPTOAuthClient(LLMClientBase):
    """
    LLM client for ChatGPT OAuth provider.

    This client:
    1. Transforms standard OpenAI chat format to ChatGPT backend Responses API format
    2. Adds required headers (Authorization, ChatGPT-Account-Id, OpenAI-Beta, OpenAI-Originator)
    3. Makes requests to chatgpt.com/backend-api/codex/responses
    4. Transforms responses back to OpenAI ChatCompletion format
    """

    MAX_RETRIES = 3
    # Transient httpx errors that are safe to retry (connection drops, transport-level failures)
    _RETRYABLE_ERRORS = (httpx.ReadError, httpx.WriteError, httpx.ConnectError, httpx.RemoteProtocolError, LLMConnectionError)

    @trace_method
    async def _get_provider_and_credentials_async(self, llm_config: LLMConfig) -> tuple[ChatGPTOAuthProvider, ChatGPTOAuthCredentials]:
        """Get the ChatGPT OAuth provider and credentials with automatic refresh if needed.

        Args:
            llm_config: The LLM configuration containing provider info.

        Returns:
            Tuple of (provider, credentials).

        Raises:
            LLMAuthenticationError: If credentials cannot be obtained.
        """
        from letta.services.provider_manager import ProviderManager

        if llm_config.provider_category != ProviderCategory.byok:
            raise ValueError("ChatGPT OAuth requires BYOK provider credentials")

        # Get provider
        provider_manager = ProviderManager()
        providers = await provider_manager.list_providers_async(
            name=llm_config.provider_name,
            actor=self.actor,
            provider_category=[ProviderCategory.byok],
        )

        if not providers:
            raise LLMAuthenticationError(
                message=f"ChatGPT OAuth provider '{llm_config.provider_name}' not found",
                code=ErrorCode.UNAUTHENTICATED,
            )

        provider: ChatGPTOAuthProvider = providers[0].cast_to_subtype()

        # Get credentials with automatic refresh (pass actor for persistence)
        creds = await provider.refresh_token_if_needed(actor=self.actor)
        if not creds:
            raise LLMAuthenticationError(
                message="Failed to obtain valid ChatGPT OAuth credentials",
                code=ErrorCode.UNAUTHENTICATED,
            )

        return provider, creds

    def _build_headers(self, creds: ChatGPTOAuthCredentials) -> Dict[str, str]:
        """Build required headers for ChatGPT backend API.

        Args:
            creds: OAuth credentials containing access_token and account_id.

        Returns:
            Dictionary of HTTP headers.
        """
        if not creds.access_token:
            raise LLMAuthenticationError(
                message="ChatGPT OAuth access_token is empty or missing",
                code=ErrorCode.UNAUTHENTICATED,
            )
        return {
            "Authorization": f"Bearer {creds.access_token}",
            "ChatGPT-Account-Id": creds.account_id,
            "OpenAI-Beta": "responses=v1",
            "OpenAI-Originator": "codex",
            "Content-Type": "application/json",
            "accept": "text/event-stream",
        }

    @trace_method
    def build_request_data(
        self,
        agent_type: AgentType,
        messages: List[PydanticMessage],
        llm_config: LLMConfig,
        tools: Optional[List[dict]] = None,
        force_tool_call: Optional[str] = None,
        requires_subsequent_tool_call: bool = False,
        tool_return_truncation_chars: Optional[int] = None,
    ) -> dict:
        """
        Build request data for ChatGPT backend API in Responses API format.

        The ChatGPT backend uses the OpenAI Responses API format:
        - `input` array instead of `messages`
        - `role: "developer"` instead of `role: "system"`
        - Structured content arrays
        """
        # Use the existing method to convert messages to Responses API format
        input_messages = PydanticMessage.to_openai_responses_dicts_from_list(
            messages,
            tool_return_truncation_chars=tool_return_truncation_chars,
        )

        # Extract system message as instructions
        instructions = None
        filtered_input = []
        for msg in input_messages:
            if msg.get("role") == "developer":
                # First developer message becomes instructions
                if instructions is None:
                    content = msg.get("content", [])
                    if isinstance(content, list) and content:
                        instructions = content[0].get("text", "")
                    elif isinstance(content, str):
                        instructions = content
                else:
                    filtered_input.append(msg)
            else:
                filtered_input.append(msg)

        # Build tool_choice
        tool_choice = None
        if tools:
            if force_tool_call is not None:
                tool_choice = {"type": "function", "name": force_tool_call}
            elif requires_subsequent_tool_call:
                tool_choice = "required"
            else:
                tool_choice = "auto"

        # Build request payload for ChatGPT backend
        data: Dict[str, Any] = {
            "model": llm_config.model,
            "input": filtered_input,
            "store": False,  # Required for stateless operation
            "stream": True,  # ChatGPT backend requires streaming
        }

        if instructions:
            data["instructions"] = instructions

        if tools:
            # Convert tools to Responses API format
            responses_tools = [
                {
                    "type": "function",
                    "name": t.get("name"),
                    "description": t.get("description"),
                    "parameters": t.get("parameters"),
                }
                for t in tools
            ]
            data["tools"] = responses_tools
            data["tool_choice"] = tool_choice

        # Note: ChatGPT backend does NOT support max_output_tokens parameter

        # Add reasoning effort for reasoning models (GPT-5.x, o-series)
        if self.is_reasoning_model(llm_config) and llm_config.reasoning_effort:
            data["reasoning"] = {
                "effort": llm_config.reasoning_effort,
                "summary": "detailed",
            }

        return data

    def _transform_response_from_chatgpt_backend(self, response_data: dict) -> dict:
        """Transform ChatGPT backend response to standard OpenAI ChatCompletion format.

        The ChatGPT backend returns responses in Responses API format.
        This method normalizes them to ChatCompletion format.

        Args:
            response_data: Raw response from ChatGPT backend.

        Returns:
            Response in OpenAI ChatCompletion format.
        """
        # If response is already in ChatCompletion format, return as-is
        if "choices" in response_data:
            return response_data

        # Extract from Responses API format
        output = response_data.get("output", [])
        message_content = ""
        tool_calls = None
        reasoning_content = ""

        for item in output:
            item_type = item.get("type")

            if item_type == "message":
                content_parts = item.get("content", [])
                for part in content_parts:
                    if part.get("type") in ("output_text", "text"):
                        message_content += part.get("text", "")
                    elif part.get("type") == "refusal":
                        message_content += part.get("refusal", "")

            elif item_type == "function_call":
                if tool_calls is None:
                    tool_calls = []
                tool_calls.append(
                    {
                        "id": item.get("call_id", item.get("id", "")),
                        "type": "function",
                        "function": {
                            "name": item.get("name", ""),
                            "arguments": item.get("arguments", ""),
                        },
                    }
                )

            elif item_type == "reasoning":
                # Capture reasoning/thinking content if present
                summary = item.get("summary", [])
                for s in summary:
                    if s.get("type") == "summary_text":
                        reasoning_content += s.get("text", "")

        # Build ChatCompletion response
        finish_reason = "stop"
        if tool_calls:
            finish_reason = "tool_calls"

        transformed = {
            "id": response_data.get("id", "chatgpt-response"),
            "object": "chat.completion",
            "created": response_data.get("created_at", 0),
            "model": response_data.get("model", ""),
            "choices": [
                {
                    "index": 0,
                    "message": {
                        "role": "assistant",
                        "content": message_content or None,
                        "tool_calls": tool_calls,
                    },
                    "finish_reason": finish_reason,
                }
            ],
            "usage": self._transform_usage(response_data.get("usage", {})),
        }

        return transformed

    def _transform_usage(self, usage: dict) -> dict:
        """Transform usage statistics from Responses API format."""
        return {
            "prompt_tokens": usage.get("input_tokens", 0),
            "completion_tokens": usage.get("output_tokens", 0),
            "total_tokens": usage.get("input_tokens", 0) + usage.get("output_tokens", 0),
        }

    @trace_method
    def request(self, request_data: dict, llm_config: LLMConfig) -> dict:
        """Synchronous request - not recommended for ChatGPT OAuth."""
        import asyncio

        return asyncio.run(self.request_async(request_data, llm_config))

    @trace_method
    async def request_async(self, request_data: dict, llm_config: LLMConfig) -> dict:
        """Make asynchronous request to ChatGPT backend API.

        Args:
            request_data: Request payload in Responses API format.
            llm_config: LLM configuration.

        Returns:
            Response data in OpenAI ChatCompletion format.
        """
        request_data = sanitize_unicode_surrogates(request_data)

        _, creds = await self._get_provider_and_credentials_async(llm_config)
        headers = self._build_headers(creds)

        endpoint = llm_config.model_endpoint or CHATGPT_CODEX_ENDPOINT

        # ChatGPT backend requires streaming, so we use client.stream() to handle SSE
        # Retry on transient network errors with exponential backoff
        for attempt in range(self.MAX_RETRIES):
            try:
                async with httpx.AsyncClient() as client:
                    async with client.stream(
                        "POST",
                        endpoint,
                        json=request_data,
                        headers=headers,
                        timeout=120.0,
                    ) as response:
                        response.raise_for_status()
                        # Accumulate SSE events into a final response
                        return await self._accumulate_sse_response(response)

            except httpx.HTTPStatusError as e:
                mapped = self._handle_http_error(e)
                if isinstance(mapped, tuple(self._RETRYABLE_ERRORS)) and attempt < self.MAX_RETRIES - 1:
                    wait = 2**attempt
                    logger.warning(
                        f"[ChatGPT] Retryable HTTP error on request (attempt {attempt + 1}/{self.MAX_RETRIES}), "
                        f"retrying in {wait}s: {type(mapped).__name__}: {mapped}"
                    )
                    await asyncio.sleep(wait)
                    continue
                raise mapped
            except httpx.TimeoutException:
                raise LLMTimeoutError(
                    message="ChatGPT backend request timed out",
                    code=ErrorCode.TIMEOUT,
                )
            except self._RETRYABLE_ERRORS as e:
                if attempt < self.MAX_RETRIES - 1:
                    wait = 2**attempt
                    logger.warning(
                        f"[ChatGPT] Transient error on request (attempt {attempt + 1}/{self.MAX_RETRIES}), "
                        f"retrying in {wait}s: {type(e).__name__}: {e}"
                    )
                    await asyncio.sleep(wait)
                    continue
                raise LLMConnectionError(
                    message=f"Failed to connect to ChatGPT backend after {self.MAX_RETRIES} attempts: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                    details={"cause": str(e.__cause__) if e.__cause__ else None, "error_type": type(e).__name__},
                )
            except httpx.RequestError as e:
                raise LLMConnectionError(
                    message=f"Failed to connect to ChatGPT backend: {str(e)}",
                    code=ErrorCode.INTERNAL_SERVER_ERROR,
                )

        # Should not be reached, but satisfy type checker
        raise LLMConnectionError(message="ChatGPT request failed after all retries", code=ErrorCode.INTERNAL_SERVER_ERROR)

    async def _accumulate_sse_response(self, response: httpx.Response) -> dict:
        """Accumulate SSE stream into a final response.

        ChatGPT backend may return SSE even for non-streaming requests.
        This method accumulates all events into a single response.

        Args:
            response: httpx Response object with SSE content.

        Returns:
            Accumulated response data.
        """
        accumulated_content = ""
        accumulated_tool_calls: List[Dict[str, Any]] = []
        model = ""
        response_id = ""
        usage = {}

        async for line in response.aiter_lines():
            if not line.startswith("data: "):
                continue

            data_str = line[6:]  # Remove "data: " prefix
            if data_str == "[DONE]":
                break

            try:
                event = json.loads(data_str)
            except json.JSONDecodeError:
                continue

            # Extract response metadata
            if not response_id and event.get("id"):
                response_id = event["id"]
            if not model and event.get("model"):
                model = event["model"]
            if event.get("usage"):
                usage = event["usage"]

            # Handle different event types
            event_type = event.get("type")

            if event_type == "response.output_item.done":
                item = event.get("item", {})
                item_type = item.get("type")

                if item_type == "message":
                    for content in item.get("content", []):
                        if content.get("type") in ("output_text", "text"):
                            accumulated_content += content.get("text", "")

                elif item_type == "function_call":
                    accumulated_tool_calls.append(
                        {
                            "id": item.get("call_id", item.get("id", "")),
                            "type": "function",
                            "function": {
                                "name": item.get("name", ""),
                                "arguments": item.get("arguments", ""),
                            },
                        }
                    )

            elif event_type == "response.content_part.delta":
                delta = event.get("delta", {})
                if delta.get("type") == "text_delta":
                    accumulated_content += delta.get("text", "")

            elif event_type == "response.done":
                # Final response event
                if event.get("response", {}).get("usage"):
                    usage = event["response"]["usage"]

        # Build final response
        finish_reason = "stop" if not accumulated_tool_calls else "tool_calls"

        return {
            "id": response_id or "chatgpt-response",
            "object": "chat.completion",
            "created": 0,
            "model": model,
            "choices": [
                {
                    "index": 0,
                    "message": {
                        "role": "assistant",
                        "content": accumulated_content or None,
                        "tool_calls": accumulated_tool_calls if accumulated_tool_calls else None,
                    },
                    "finish_reason": finish_reason,
                }
            ],
            "usage": self._transform_usage(usage),
        }

    @trace_method
    async def request_embeddings(
        self,
        texts: List[str],
        embedding_config,
    ) -> List[List[float]]:
        """ChatGPT backend does not support embeddings."""
        raise NotImplementedError("ChatGPT OAuth does not support embeddings")

    @trace_method
    async def convert_response_to_chat_completion(
        self,
        response_data: dict,
        input_messages: List[PydanticMessage],
        llm_config: LLMConfig,
    ) -> ChatCompletionResponse:
        """Convert response to ChatCompletionResponse.

        Args:
            response_data: Response data (already in ChatCompletion format).
            input_messages: Original input messages.
            llm_config: LLM configuration.

        Returns:
            ChatCompletionResponse object.
        """
        # Response should already be in ChatCompletion format after transformation
        return ChatCompletionResponse(**response_data)

    def extract_usage_statistics(self, response_data: dict | None, llm_config: LLMConfig) -> LettaUsageStatistics:
        """Extract usage statistics from ChatGPT OAuth response and return as LettaUsageStatistics."""
        if not response_data:
            return LettaUsageStatistics()

        usage = response_data.get("usage")
        if not usage:
            return LettaUsageStatistics()

        prompt_tokens = usage.get("prompt_tokens") or 0
        completion_tokens = usage.get("completion_tokens") or 0
        total_tokens = usage.get("total_tokens") or (prompt_tokens + completion_tokens)

        return LettaUsageStatistics(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=total_tokens,
        )

    @trace_method
    async def stream_async(
        self,
        request_data: dict,
        llm_config: LLMConfig,
    ) -> AsyncStreamWrapper:
        """Stream response from ChatGPT backend.

        Note: ChatGPT backend uses SSE by default. This returns a custom
        async generator that yields ResponseStreamEvent objects compatible
        with the OpenAI SDK.

        Args:
            request_data: Request payload.
            llm_config: LLM configuration.

        Returns:
            Async generator yielding ResponseStreamEvent objects.
        """
        request_data = sanitize_unicode_surrogates(request_data)

        _, creds = await self._get_provider_and_credentials_async(llm_config)
        headers = self._build_headers(creds)

        endpoint = llm_config.model_endpoint or CHATGPT_CODEX_ENDPOINT

        async def stream_generator():
            # Track output item index for proper event construction
            output_index = 0
            # Track sequence_number in case backend doesn't provide it
            # (OpenAI SDK expects incrementing sequence numbers starting at 0)
            sequence_counter = 0
            # Track whether we've yielded any events — once we have, we can't
            # transparently retry because the caller has already consumed partial data.
            has_yielded = False

            for attempt in range(self.MAX_RETRIES):
                try:
                    async with httpx.AsyncClient() as client:
                        async with client.stream(
                            "POST",
                            endpoint,
                            json=request_data,
                            headers=headers,
                            timeout=120.0,
                        ) as response:
                            # Check for error status
                            if response.status_code != 200:
                                error_body = await response.aread()
                                logger.error(f"ChatGPT SSE error: {response.status_code} - {error_body}")
                                raise self._handle_http_error_from_status(response.status_code, error_body.decode())

                            async for line in response.aiter_lines():
                                if not line or not line.startswith("data: "):
                                    continue

                                data_str = line[6:]
                                if data_str == "[DONE]":
                                    break

                                try:
                                    raw_event = json.loads(data_str)
                                    event_type = raw_event.get("type")

                                    # Check for error events from the API (context window, rate limit, etc.)
                                    if event_type == "error":
                                        logger.error(f"ChatGPT SSE error event: {json.dumps(raw_event, default=str)[:1000]}")
                                        raise self._handle_sse_error_event(raw_event)

                                    # Check for response.failed or response.incomplete events
                                    if event_type in ("response.failed", "response.incomplete"):
                                        logger.error(f"ChatGPT SSE {event_type} event: {json.dumps(raw_event, default=str)[:1000]}")
                                        resp_obj = raw_event.get("response", {})
                                        error_info = resp_obj.get("error", {})
                                        if error_info:
                                            raise self._handle_sse_error_event({"error": error_info, "type": event_type})
                                        else:
                                            raise LLMBadRequestError(
                                                message=f"ChatGPT request failed with status '{event_type}' (no error details provided)",
                                                code=ErrorCode.INTERNAL_SERVER_ERROR,
                                            )

                                    # Use backend-provided sequence_number if available, else use counter
                                    # This ensures proper ordering even if backend doesn't provide it
                                    if "sequence_number" not in raw_event:
                                        raw_event["sequence_number"] = sequence_counter
                                    sequence_counter = raw_event["sequence_number"] + 1

                                    # Track output index for output_item.added events
                                    if event_type == "response.output_item.added":
                                        output_index = raw_event.get("output_index", output_index)

                                    # Convert to OpenAI SDK ResponseStreamEvent
                                    sdk_event = self._convert_to_sdk_event(raw_event, output_index)
                                    if sdk_event:
                                        yield sdk_event
                                        has_yielded = True

                                except json.JSONDecodeError:
                                    logger.warning(f"Failed to parse SSE event: {data_str[:100]}")
                                    continue

                    # Stream completed successfully
                    return

                except self._RETRYABLE_ERRORS as e:
                    if has_yielded or attempt >= self.MAX_RETRIES - 1:
                        # Already yielded partial data or exhausted retries — must propagate
                        raise
                    wait = 2**attempt
                    logger.warning(
                        f"[ChatGPT] Transient error on stream (attempt {attempt + 1}/{self.MAX_RETRIES}), "
                        f"retrying in {wait}s: {type(e).__name__}: {e}"
                    )
                    await asyncio.sleep(wait)

        # Wrap the async generator in AsyncStreamWrapper to provide context manager protocol
        return AsyncStreamWrapper(stream_generator())

    def _convert_to_sdk_event(
        self,
        raw_event: dict,
        output_index: int = 0,
    ) -> Optional[ResponseStreamEvent]:
        """Convert raw ChatGPT backend SSE event to OpenAI SDK ResponseStreamEvent.

        Uses model_construct() to bypass validation since ChatGPT backend doesn't
        provide all fields required by OpenAI SDK (e.g., sequence_number).

        Args:
            raw_event: Raw SSE event data from ChatGPT backend.
            output_index: Current output item index.

        Returns:
            OpenAI SDK ResponseStreamEvent or None if event type not handled.
        """
        event_type = raw_event.get("type")
        response_id = raw_event.get("response_id", "")
        seq_num = raw_event.get("sequence_number", 0)

        # response.created -> ResponseCreatedEvent
        if event_type == "response.created":
            response_data = raw_event.get("response", {})
            return ResponseCreatedEvent.model_construct(
                type="response.created",
                sequence_number=seq_num,
                response=Response.model_construct(
                    id=response_data.get("id", response_id),
                    created_at=response_data.get("created_at", 0),
                    model=response_data.get("model", ""),
                    object="response",
                    output=[],
                    status=response_data.get("status", "in_progress"),
                    parallel_tool_calls=response_data.get("parallel_tool_calls", True),
                ),
            )

        # response.in_progress -> ResponseInProgressEvent
        elif event_type == "response.in_progress":
            response_data = raw_event.get("response", {})
            return ResponseInProgressEvent.model_construct(
                type="response.in_progress",
                sequence_number=seq_num,
                response=Response.model_construct(
                    id=response_data.get("id", response_id),
                    created_at=response_data.get("created_at", 0),
                    model=response_data.get("model", ""),
                    object="response",
                    output=[],
                    status="in_progress",
                    parallel_tool_calls=response_data.get("parallel_tool_calls", True),
                ),
            )

        # response.output_item.added -> ResponseOutputItemAddedEvent
        elif event_type == "response.output_item.added":
            item_data = raw_event.get("item", {})
            item_type = item_data.get("type")
            idx = raw_event.get("output_index", output_index)

            if item_type == "message":
                item = ResponseOutputMessage.model_construct(
                    id=item_data.get("id", ""),
                    type="message",
                    role=item_data.get("role", "assistant"),
                    content=[],
                    status=item_data.get("status", "in_progress"),
                )
            elif item_type == "function_call":
                item = ResponseFunctionToolCall.model_construct(
                    id=item_data.get("id", ""),
                    type="function_call",
                    call_id=item_data.get("call_id", ""),
                    name=item_data.get("name", ""),
                    arguments=item_data.get("arguments", ""),
                    status=item_data.get("status", "in_progress"),
                )
            elif item_type == "reasoning":
                # Reasoning item for o-series, GPT-5 models
                item = ResponseReasoningItem.model_construct(
                    id=item_data.get("id", ""),
                    type="reasoning",
                    summary=item_data.get("summary", []),
                    status=item_data.get("status", "in_progress"),
                )
            else:
                # Unknown item type, skip
                return None

            return ResponseOutputItemAddedEvent.model_construct(
                type="response.output_item.added",
                sequence_number=seq_num,
                output_index=idx,
                item=item,
            )

        # response.content_part.added -> ResponseContentPartAddedEvent
        elif event_type == "response.content_part.added":
            part_data = raw_event.get("part", {})
            return ResponseContentPartAddedEvent.model_construct(
                type="response.content_part.added",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                content_index=raw_event.get("content_index", 0),
                part=ResponseOutputText.model_construct(
                    type="output_text",
                    text=part_data.get("text", ""),
                    annotations=[],
                ),
            )

        # response.output_text.delta -> ResponseTextDeltaEvent
        # Note: OpenAI SDK uses "response.output_text.delta" (matching ChatGPT backend)
        elif event_type == "response.output_text.delta":
            return ResponseTextDeltaEvent.model_construct(
                type="response.output_text.delta",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                content_index=raw_event.get("content_index", 0),
                delta=raw_event.get("delta", ""),
            )

        # response.output_text.done -> ResponseTextDoneEvent
        elif event_type == "response.output_text.done":
            return ResponseTextDoneEvent.model_construct(
                type="response.output_text.done",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                content_index=raw_event.get("content_index", 0),
                text=raw_event.get("text", ""),
            )

        # response.function_call_arguments.delta -> ResponseFunctionCallArgumentsDeltaEvent
        elif event_type == "response.function_call_arguments.delta":
            return ResponseFunctionCallArgumentsDeltaEvent.model_construct(
                type="response.function_call_arguments.delta",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                call_id=raw_event.get("call_id", ""),
                delta=raw_event.get("delta", ""),
            )

        # response.function_call_arguments.done -> ResponseFunctionCallArgumentsDoneEvent
        elif event_type == "response.function_call_arguments.done":
            return ResponseFunctionCallArgumentsDoneEvent.model_construct(
                type="response.function_call_arguments.done",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                call_id=raw_event.get("call_id", ""),
                arguments=raw_event.get("arguments", ""),
            )

        # response.content_part.done -> ResponseContentPartDoneEvent
        elif event_type == "response.content_part.done":
            part_data = raw_event.get("part", {})
            return ResponseContentPartDoneEvent.model_construct(
                type="response.content_part.done",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                content_index=raw_event.get("content_index", 0),
                part=ResponseOutputText.model_construct(
                    type="output_text",
                    text=part_data.get("text", ""),
                    annotations=[],
                ),
            )

        # response.output_item.done -> ResponseOutputItemDoneEvent
        elif event_type == "response.output_item.done":
            item_data = raw_event.get("item", {})
            item_type = item_data.get("type")
            idx = raw_event.get("output_index", output_index)

            if item_type == "message":
                # Build content from item data
                content_list = []
                for c in item_data.get("content", []):
                    if c.get("type") in ("output_text", "text"):
                        content_list.append(
                            ResponseOutputText.model_construct(
                                type="output_text",
                                text=c.get("text", ""),
                                annotations=[],
                            )
                        )
                item = ResponseOutputMessage.model_construct(
                    id=item_data.get("id", ""),
                    type="message",
                    role=item_data.get("role", "assistant"),
                    content=content_list,
                    status=item_data.get("status", "completed"),
                )
            elif item_type == "function_call":
                item = ResponseFunctionToolCall.model_construct(
                    id=item_data.get("id", ""),
                    type="function_call",
                    call_id=item_data.get("call_id", ""),
                    name=item_data.get("name", ""),
                    arguments=item_data.get("arguments", ""),
                    status=item_data.get("status", "completed"),
                )
            elif item_type == "reasoning":
                # Build summary from item data
                summary_list = item_data.get("summary", [])
                item = ResponseReasoningItem.model_construct(
                    id=item_data.get("id", ""),
                    type="reasoning",
                    summary=summary_list,
                    status=item_data.get("status", "completed"),
                )
            else:
                return None

            return ResponseOutputItemDoneEvent.model_construct(
                type="response.output_item.done",
                sequence_number=seq_num,
                output_index=idx,
                item=item,
            )

        # response.completed or response.done -> ResponseCompletedEvent
        elif event_type in ("response.completed", "response.done"):
            response_data = raw_event.get("response", {})

            # Build output items from response data
            output_items = []
            for out in response_data.get("output", []):
                out_type = out.get("type")
                if out_type == "message":
                    content_list = []
                    for c in out.get("content", []):
                        if c.get("type") in ("output_text", "text"):
                            content_list.append(
                                ResponseOutputText.model_construct(
                                    type="output_text",
                                    text=c.get("text", ""),
                                    annotations=[],
                                )
                            )
                    output_items.append(
                        ResponseOutputMessage.model_construct(
                            id=out.get("id", ""),
                            type="message",
                            role=out.get("role", "assistant"),
                            content=content_list,
                            status=out.get("status", "completed"),
                        )
                    )
                elif out_type == "function_call":
                    output_items.append(
                        ResponseFunctionToolCall.model_construct(
                            id=out.get("id", ""),
                            type="function_call",
                            call_id=out.get("call_id", ""),
                            name=out.get("name", ""),
                            arguments=out.get("arguments", ""),
                            status=out.get("status", "completed"),
                        )
                    )

            return ResponseCompletedEvent.model_construct(
                type="response.completed",
                sequence_number=seq_num,
                response=Response.model_construct(
                    id=response_data.get("id", response_id),
                    created_at=response_data.get("created_at", 0),
                    model=response_data.get("model", ""),
                    object="response",
                    output=output_items,
                    status=response_data.get("status", "completed"),
                    parallel_tool_calls=response_data.get("parallel_tool_calls", True),
                    usage=response_data.get("usage"),
                ),
            )

        # Reasoning events (for o-series, GPT-5 models)
        # response.reasoning_summary_part.added -> ResponseReasoningSummaryPartAddedEvent
        elif event_type == "response.reasoning_summary_part.added":
            part_data = raw_event.get("part", {})
            # Use a simple dict for Part since we use model_construct
            part = {"text": part_data.get("text", ""), "type": part_data.get("type", "summary_text")}
            return ResponseReasoningSummaryPartAddedEvent.model_construct(
                type="response.reasoning_summary_part.added",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                summary_index=raw_event.get("summary_index", 0),
                part=part,
            )

        # response.reasoning_summary_text.delta -> ResponseReasoningSummaryTextDeltaEvent
        elif event_type == "response.reasoning_summary_text.delta":
            return ResponseReasoningSummaryTextDeltaEvent.model_construct(
                type="response.reasoning_summary_text.delta",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                summary_index=raw_event.get("summary_index", 0),
                delta=raw_event.get("delta", ""),
            )

        # response.reasoning_summary_text.done -> ResponseReasoningSummaryTextDoneEvent
        elif event_type == "response.reasoning_summary_text.done":
            return ResponseReasoningSummaryTextDoneEvent.model_construct(
                type="response.reasoning_summary_text.done",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                summary_index=raw_event.get("summary_index", 0),
                text=raw_event.get("text", ""),
            )

        # response.reasoning_summary_part.done -> ResponseReasoningSummaryPartDoneEvent
        elif event_type == "response.reasoning_summary_part.done":
            part_data = raw_event.get("part", {})
            part = {"text": part_data.get("text", ""), "type": part_data.get("type", "summary_text")}
            return ResponseReasoningSummaryPartDoneEvent.model_construct(
                type="response.reasoning_summary_part.done",
                sequence_number=seq_num,
                item_id=raw_event.get("item_id", ""),
                output_index=raw_event.get("output_index", output_index),
                summary_index=raw_event.get("summary_index", 0),
                part=part,
            )

        # Unhandled event types
        logger.warning(f"Unhandled ChatGPT SSE event type: {event_type}")
        return None

    @staticmethod
    def _is_upstream_connection_error(error_body: str) -> bool:
        """Check if an error body indicates an upstream connection/proxy failure."""
        lower = error_body.lower()
        return "upstream connect error" in lower or "reset before headers" in lower or "connection termination" in lower

    def _handle_http_error_from_status(self, status_code: int, error_body: str) -> Exception:
        """Create appropriate exception from HTTP status code.

        Args:
            status_code: HTTP status code.
            error_body: Error response body.

        Returns:
            Appropriate LLM exception.
        """
        if status_code == 401:
            return LLMAuthenticationError(
                message=f"ChatGPT authentication failed: {error_body}",
                code=ErrorCode.UNAUTHENTICATED,
            )
        elif status_code == 429:
            return LLMRateLimitError(
                message=f"ChatGPT rate limit exceeded: {error_body}",
                code=ErrorCode.RATE_LIMIT_EXCEEDED,
            )
        elif status_code == 502 or (status_code >= 500 and self._is_upstream_connection_error(error_body)):
            return LLMConnectionError(
                message=f"ChatGPT upstream connection error: {error_body}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
            )
        elif status_code >= 500:
            return LLMServerError(
                message=f"ChatGPT API error: {error_body}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
            )
        else:
            return LLMBadRequestError(
                message=f"ChatGPT request failed ({status_code}): {error_body}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
            )

    def is_reasoning_model(self, llm_config: LLMConfig) -> bool:
        """Check if model is a reasoning model.

        Args:
            llm_config: LLM configuration.

        Returns:
            True if model supports extended reasoning.
        """
        model = llm_config.model.lower()
        return "o1" in model or "o3" in model or "o4" in model or "gpt-5" in model

    @trace_method
    def handle_llm_error(self, e: Exception, llm_config: Optional[LLMConfig] = None) -> Exception:
        """Map ChatGPT-specific errors to common LLMError types.

        Args:
            e: Original exception.
            llm_config: Optional LLM config to determine if this is a BYOK key.

        Returns:
            Mapped LLMError subclass.
        """
        is_byok = (llm_config.provider_category == ProviderCategory.byok) if llm_config else None

        # Already a typed LLM/Letta error (e.g. from SSE error handling) — pass through
        if isinstance(e, LettaError):
            return e

        if isinstance(e, httpx.HTTPStatusError):
            return self._handle_http_error(e, is_byok=is_byok)

        # Handle httpx network errors which can occur during streaming
        # when the connection is unexpectedly closed while reading/writing
        if isinstance(e, (httpx.ReadError, httpx.WriteError, httpx.ConnectError)):
            logger.warning(f"[ChatGPT] Network error during streaming: {type(e).__name__}: {e}")
            return LLMConnectionError(
                message=f"Network error during ChatGPT streaming: {str(e)}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                details={"cause": str(e.__cause__) if e.__cause__ else None, "error_type": type(e).__name__, "is_byok": is_byok},
            )

        return super().handle_llm_error(e, llm_config=llm_config)

    def _handle_http_error(self, e: httpx.HTTPStatusError, is_byok: bool | None = None) -> Exception:
        """Handle HTTP status errors from ChatGPT backend.

        Args:
            e: HTTP status error.
            is_byok: Whether the request used a BYOK key.

        Returns:
            Appropriate LLMError subclass.
        """
        status_code = e.response.status_code
        error_text = str(e)

        try:
            error_json = e.response.json()
            error_message = error_json.get("error", {}).get("message", error_text)
        except Exception:
            error_message = error_text

        if status_code == 401:
            return LLMAuthenticationError(
                message=f"ChatGPT authentication failed: {error_message}",
                code=ErrorCode.UNAUTHENTICATED,
                details={"is_byok": is_byok},
            )
        elif status_code == 429:
            return LLMRateLimitError(
                message=f"ChatGPT rate limit exceeded: {error_message}",
                code=ErrorCode.RATE_LIMIT_EXCEEDED,
                details={"is_byok": is_byok},
            )
        elif status_code == 400:
            if "context" in error_message.lower() or "token" in error_message.lower():
                return ContextWindowExceededError(
                    message=f"ChatGPT context window exceeded: {error_message}",
                    details={"is_byok": is_byok},
                )
            return LLMBadRequestError(
                message=f"ChatGPT bad request: {error_message}",
                code=ErrorCode.INVALID_ARGUMENT,
                details={"is_byok": is_byok},
            )
        elif status_code == 502 or (status_code >= 500 and self._is_upstream_connection_error(error_message)):
            return LLMConnectionError(
                message=f"ChatGPT upstream connection error: {error_message}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                details={"is_byok": is_byok},
            )
        elif status_code >= 500:
            return LLMServerError(
                message=f"ChatGPT API error: {error_message}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                details={"is_byok": is_byok},
            )
        else:
            return LLMBadRequestError(
                message=f"ChatGPT request failed ({status_code}): {error_message}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
                details={"is_byok": is_byok},
            )

    def _handle_sse_error_event(self, raw_event: dict) -> Exception:
        """Create appropriate exception from an SSE error or response.failed event.

        The ChatGPT backend can return errors as SSE events within a 200 OK stream,
        e.g. {"type": "error", "error": {"type": "invalid_request_error",
        "code": "context_length_exceeded", "message": "..."}}.

        Args:
            raw_event: Raw SSE event data containing an error.

        Returns:
            Appropriate LLM exception.
        """
        error_obj = raw_event.get("error", {})
        if isinstance(error_obj, str):
            error_message = error_obj
            error_code = None
        else:
            error_message = error_obj.get("message", "Unknown ChatGPT SSE error")
            error_code = error_obj.get("code") or None

        if error_code == "context_length_exceeded":
            return ContextWindowExceededError(
                message=f"ChatGPT context window exceeded: {error_message}",
            )
        elif error_code == "rate_limit_exceeded":
            return LLMRateLimitError(
                message=f"ChatGPT rate limit exceeded: {error_message}",
                code=ErrorCode.RATE_LIMIT_EXCEEDED,
            )
        elif error_code == "authentication_error":
            return LLMAuthenticationError(
                message=f"ChatGPT authentication failed: {error_message}",
                code=ErrorCode.UNAUTHENTICATED,
            )
        elif error_code == "server_error":
            return LLMServerError(
                message=f"ChatGPT API error: {error_message}",
                code=ErrorCode.INTERNAL_SERVER_ERROR,
            )
        else:
            return LLMBadRequestError(
                message=f"ChatGPT SSE error ({error_code or 'unknown'}): {error_message}",
                code=ErrorCode.INVALID_ARGUMENT,
            )



from typing import List, Literal, Optional


def run_code(code: str, language: Literal["python", "js", "ts", "r", "java"]) -> str:
    """
    Run code in a sandbox. Supports Python, Javascript, Typescript, R, and Java.

    Args:
        code (str): The code to run.
        language (Literal["python", "js", "ts", "r", "java"]): The language of the code.
    Returns:
        str: The output of the code, the stdout, the stderr, and error traces (if any).
    """

    raise NotImplementedError("This is only available on the latest agent architecture. Please contact the Letta team.")


def run_code_with_tools(code: str) -> str:
    """
    Run code with access to the tools of the agent. Only support python. You can directly invoke the tools of the agent in the code.
    Args:
        code (str): The python code to run.
    Returns:
        str: The output of the code, the stdout, the stderr, and error traces (if any).
    """

    raise NotImplementedError("This is only available on the latest agent architecture. Please contact the Letta team.")


async def web_search(
    query: str,
    num_results: int = 10,
    category: Optional[
        Literal["company", "research paper", "news", "pdf", "github", "tweet", "personal site", "linkedin profile", "financial report"]
    ] = None,
    include_text: bool = False,
    include_domains: Optional[List[str]] = None,
    exclude_domains: Optional[List[str]] = None,
    start_published_date: Optional[str] = None,
    end_published_date: Optional[str] = None,
    user_location: Optional[str] = None,
) -> str:
    """
    Search the web using Exa's AI-powered search engine and retrieve relevant content.

    Examples:
    web_search("Tesla Q1 2025 earnings report", num_results=5, category="financial report")
    web_search("Latest research in large language models", category="research paper", include_domains=["arxiv.org", "paperswithcode.com"])
    web_search("Letta API documentation core_memory_append", num_results=3)

    Args:
        query (str): The search query to find relevant web content.
        num_results (int, optional): Number of results to return (1-100). Defaults to 10.
        category (Optional[Literal], optional): Focus search on specific content types. Defaults to None.
        include_text (bool, optional): Whether to retrieve full page content. Defaults to False (only returns summary and highlights, since the full text usually will overflow the context window).
        include_domains (Optional[List[str]], optional): List of domains to include in search results. Defaults to None.
        exclude_domains (Optional[List[str]], optional): List of domains to exclude from search results. Defaults to None.
        start_published_date (Optional[str], optional): Only return content published after this date (ISO format). Defaults to None.
        end_published_date (Optional[str], optional): Only return content published before this date (ISO format). Defaults to None.
        user_location (Optional[str], optional): Two-letter country code for localized results (e.g., "US"). Defaults to None.

    Returns:
        str: A JSON-encoded string containing search results with title, URL, content, highlights, and summary.
    """
    raise NotImplementedError("This is only available on the latest agent architecture. Please contact the Letta team.")


async def fetch_webpage(url: str) -> str:
    """
    Fetch a webpage and convert it to markdown/text format using Exa API (if available) or trafilatura/readability.

    Args:
        url: The URL of the webpage to fetch and convert

    Returns:
        String containing the webpage content in markdown/text format
    """
    raise NotImplementedError("This is only available on the latest agent architecture. Please contact the Letta team.")



import asyncio
import copy
import difflib
import hashlib
import inspect
import io
import os
import pickle
import platform
import random
import re
import subprocess
import sys
import uuid
from collections.abc import Coroutine
from contextlib import contextmanager
from datetime import datetime, timezone
from functools import wraps
from logging import Logger
from typing import Any, Callable, Optional, Union, _GenericAlias, get_args, get_origin, get_type_hints  # type: ignore[attr-defined]
from urllib.parse import urljoin, urlparse

import demjson3 as demjson
from pathvalidate import sanitize_filename as pathvalidate_sanitize_filename
from sqlalchemy import text

import letta
from letta.constants import (
    CORE_MEMORY_HUMAN_CHAR_LIMIT,
    CORE_MEMORY_PERSONA_CHAR_LIMIT,
    DEFAULT_CORE_MEMORY_SOURCE_CHAR_LIMIT,
    DEFAULT_MAX_FILES_OPEN,
    ERROR_MESSAGE_PREFIX,
    FILE_IS_TRUNCATED_WARNING,
    LETTA_DIR,
    MAX_FILENAME_LENGTH,
    TOOL_CALL_ID_MAX_LEN,
)
from letta.helpers.json_helpers import json_dumps, json_loads
from letta.log import get_logger
from letta.otel.tracing import log_attributes, trace_method
from letta.schemas.openai.chat_completion_response import ChatCompletionResponse
from letta.server.rest_api.dependencies import HeaderParams

logger = get_logger(__name__)


DEBUG = False
if "LOG_LEVEL" in os.environ:
    if os.environ["LOG_LEVEL"] == "DEBUG":
        DEBUG = True


ADJECTIVE_BANK = [
    "beautiful",
    "gentle",
    "angry",
    "vivacious",
    "grumpy",
    "luxurious",
    "fierce",
    "delicate",
    "fluffy",
    "radiant",
    "elated",
    "magnificent",
    "sassy",
    "ecstatic",
    "lustrous",
    "gleaming",
    "sorrowful",
    "majestic",
    "proud",
    "dynamic",
    "energetic",
    "mysterious",
    "loyal",
    "brave",
    "decisive",
    "frosty",
    "cheerful",
    "adorable",
    "melancholy",
    "vibrant",
    "elegant",
    "gracious",
    "inquisitive",
    "opulent",
    "peaceful",
    "rebellious",
    "scintillating",
    "dazzling",
    "whimsical",
    "impeccable",
    "meticulous",
    "resilient",
    "charming",
    "vivacious",
    "creative",
    "intuitive",
    "compassionate",
    "innovative",
    "enthusiastic",
    "tremendous",
    "effervescent",
    "tenacious",
    "fearless",
    "sophisticated",
    "witty",
    "optimistic",
    "exquisite",
    "sincere",
    "generous",
    "kindhearted",
    "serene",
    "amiable",
    "adventurous",
    "bountiful",
    "courageous",
    "diligent",
    "exotic",
    "grateful",
    "harmonious",
    "imaginative",
    "jubilant",
    "keen",
    "luminous",
    "nurturing",
    "outgoing",
    "passionate",
    "quaint",
    "resourceful",
    "sturdy",
    "tactful",
    "unassuming",
    "versatile",
    "wondrous",
    "youthful",
    "zealous",
    "ardent",
    "benevolent",
    "capricious",
    "dedicated",
    "empathetic",
    "fabulous",
    "gregarious",
    "humble",
    "intriguing",
    "jovial",
    "kind",
    "lovable",
    "mindful",
    "noble",
    "original",
    "pleasant",
    "quixotic",
    "reliable",
    "spirited",
    "tranquil",
    "unique",
    "venerable",
    "warmhearted",
    "xenodochial",
    "yearning",
    "zesty",
    "amusing",
    "blissful",
    "calm",
    "daring",
    "enthusiastic",
    "faithful",
    "graceful",
    "honest",
    "incredible",
    "joyful",
    "kind",
    "lovely",
    "merry",
    "noble",
    "optimistic",
    "peaceful",
    "quirky",
    "respectful",
    "sweet",
    "trustworthy",
    "understanding",
    "vibrant",
    "witty",
    "xenial",
    "youthful",
    "zealous",
    "ambitious",
    "brilliant",
    "careful",
    "devoted",
    "energetic",
    "friendly",
    "glorious",
    "humorous",
    "intelligent",
    "jovial",
    "knowledgeable",
    "loyal",
    "modest",
    "nice",
    "obedient",
    "patient",
    "quiet",
    "resilient",
    "selfless",
    "tolerant",
    "unique",
    "versatile",
    "warm",
    "xerothermic",
    "yielding",
    "zestful",
    "amazing",
    "bold",
    "charming",
    "determined",
    "exciting",
    "funny",
    "happy",
    "imaginative",
    "jolly",
    "keen",
    "loving",
    "magnificent",
    "nifty",
    "outstanding",
    "polite",
    "quick",
    "reliable",
    "sincere",
    "thoughtful",
    "unusual",
    "valuable",
    "wonderful",
    "xenodochial",
    "zealful",
    "admirable",
    "bright",
    "clever",
    "dedicated",
    "extraordinary",
    "generous",
    "hardworking",
    "inspiring",
    "jubilant",
    "kindhearted",
    "lively",
    "miraculous",
    "neat",
    "openminded",
    "passionate",
    "remarkable",
    "stunning",
    "truthful",
    "upbeat",
    "vivacious",
    "welcoming",
    "yare",
    "zealous",
]

NOUN_BANK = [
    "lizard",
    "firefighter",
    "banana",
    "castle",
    "dolphin",
    "elephant",
    "forest",
    "giraffe",
    "harbor",
    "iceberg",
    "jewelry",
    "kangaroo",
    "library",
    "mountain",
    "notebook",
    "orchard",
    "penguin",
    "quilt",
    "rainbow",
    "squirrel",
    "teapot",
    "umbrella",
    "volcano",
    "waterfall",
    "xylophone",
    "yacht",
    "zebra",
    "apple",
    "butterfly",
    "caterpillar",
    "dragonfly",
    "elephant",
    "flamingo",
    "gorilla",
    "hippopotamus",
    "iguana",
    "jellyfish",
    "koala",
    "lemur",
    "mongoose",
    "nighthawk",
    "octopus",
    "panda",
    "quokka",
    "rhinoceros",
    "salamander",
    "tortoise",
    "unicorn",
    "vulture",
    "walrus",
    "xenopus",
    "yak",
    "zebu",
    "asteroid",
    "balloon",
    "compass",
    "dinosaur",
    "eagle",
    "firefly",
    "galaxy",
    "hedgehog",
    "island",
    "jaguar",
    "kettle",
    "lion",
    "mammoth",
    "nucleus",
    "owl",
    "pumpkin",
    "quasar",
    "reindeer",
    "snail",
    "tiger",
    "universe",
    "vampire",
    "wombat",
    "xerus",
    "yellowhammer",
    "zeppelin",
    "alligator",
    "buffalo",
    "cactus",
    "donkey",
    "emerald",
    "falcon",
    "gazelle",
    "hamster",
    "icicle",
    "jackal",
    "kitten",
    "leopard",
    "mushroom",
    "narwhal",
    "opossum",
    "peacock",
    "quail",
    "rabbit",
    "scorpion",
    "toucan",
    "urchin",
    "viper",
    "wolf",
    "xray",
    "yucca",
    "zebu",
    "acorn",
    "biscuit",
    "cupcake",
    "daisy",
    "eyeglasses",
    "frisbee",
    "goblin",
    "hamburger",
    "icicle",
    "jackfruit",
    "kaleidoscope",
    "lighthouse",
    "marshmallow",
    "nectarine",
    "obelisk",
    "pancake",
    "quicksand",
    "raspberry",
    "spinach",
    "truffle",
    "umbrella",
    "volleyball",
    "walnut",
    "xylophonist",
    "yogurt",
    "zucchini",
    "asterisk",
    "blackberry",
    "chimpanzee",
    "dumpling",
    "espresso",
    "fireplace",
    "gnome",
    "hedgehog",
    "illustration",
    "jackhammer",
    "kumquat",
    "lemongrass",
    "mandolin",
    "nugget",
    "ostrich",
    "parakeet",
    "quiche",
    "racquet",
    "seashell",
    "tadpole",
    "unicorn",
    "vaccination",
    "wolverine",
    "yam",
    "zeppelin",
    "accordion",
    "broccoli",
    "carousel",
    "daffodil",
    "eggplant",
    "flamingo",
    "grapefruit",
    "harpsichord",
    "impression",
    "jackrabbit",
    "kitten",
    "llama",
    "mandarin",
    "nachos",
    "obelisk",
    "papaya",
    "quokka",
    "rooster",
    "sunflower",
    "turnip",
    "ukulele",
    "viper",
    "waffle",
    "xylograph",
    "yeti",
    "zephyr",
    "abacus",
    "blueberry",
    "crocodile",
    "dandelion",
    "echidna",
    "fig",
    "giraffe",
    "hamster",
    "iguana",
    "jackal",
    "kiwi",
    "lobster",
    "marmot",
    "noodle",
    "octopus",
    "platypus",
    "quail",
    "raccoon",
    "starfish",
    "tulip",
    "urchin",
    "vampire",
    "walrus",
    "xylophone",
    "yak",
    "zebra",
]


def smart_urljoin(base_url: str, relative_url: str) -> str:
    """urljoin is stupid and wants a trailing / at the end of the endpoint address, or it will chop the suffix off"""
    if not base_url.endswith("/"):
        base_url += "/"
    return urljoin(base_url, relative_url)


def get_tool_call_id() -> str:
    # TODO(sarah) make this a slug-style string?
    # e.g. OpenAI: "call_xlIfzR1HqAW7xJPa3ExJSg3C"
    # or similar to agents: "call-xlIfzR1HqAW7xJPa3ExJSg3C"
    return str(uuid.uuid4())[:TOOL_CALL_ID_MAX_LEN]


# Pattern for valid tool_call_id (required by Anthropic: ^[a-zA-Z0-9_-]+$)
TOOL_CALL_ID_PATTERN = re.compile(r"^[a-zA-Z0-9_-]+$")


def sanitize_tool_call_id(tool_id: str) -> str:
    """Ensure tool_call_id matches cross-provider requirements:
    - Anthropic: pattern ^[a-zA-Z0-9_-]+$
    - OpenAI: max length 29 characters

    Some models (e.g. Kimi via OpenRouter) generate IDs like 'Read:93' which
    contain invalid characters. This sanitizes them for cross-provider compatibility.
    """
    # Replace invalid characters with underscores
    if not TOOL_CALL_ID_PATTERN.match(tool_id):
        tool_id = re.sub(r"[^a-zA-Z0-9_-]", "_", tool_id)
    # Truncate to max length
    return tool_id[:TOOL_CALL_ID_MAX_LEN]


def assistant_function_to_tool(assistant_message: dict) -> dict:
    assert "function_call" in assistant_message
    new_msg = copy.deepcopy(assistant_message)
    function_call = new_msg.pop("function_call")
    new_msg["tool_calls"] = [
        {
            "id": get_tool_call_id(),
            "type": "function",
            "function": function_call,
        }
    ]
    return new_msg


def is_optional_type(hint):
    """Check if the type hint is an Optional type."""
    if isinstance(hint, _GenericAlias):
        return hint.__origin__ is Union and type(None) in hint.__args__
    return False


def enforce_types(func):
    """Enforces that values passed in match the expected types.
        Technically will handle coroutines as well.

    TODO (cliandy): use stricter pydantic fields
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Get type hints, excluding the return type hint
        hints = {k: v for k, v in get_type_hints(func).items() if k != "return"}

        # Get the function's argument names
        arg_names = inspect.getfullargspec(func).args

        # Pair each argument with its corresponding type hint
        args_with_hints = dict(zip(arg_names[1:], args[1:], strict=False))  # Skipping 'self'

        # Function to check if a value matches a given type hint
        def matches_type(value, hint):
            origin = get_origin(hint)
            args = get_args(hint)

            if origin is Union:  # Handle Union types (including Optional)
                return any(matches_type(value, arg) for arg in args)
            elif hasattr(hint, "__class__") and hint.__class__.__name__ == "UnionType":  # Handle Python 3.10+ X | Y syntax
                return any(matches_type(value, arg) for arg in args)
            elif origin is list and isinstance(value, list):  # Handle List[T]
                element_type = args[0] if args else None
                return all(isinstance(v, element_type) for v in value) if element_type else True
            elif origin is not None and (
                str(origin).endswith("Literal") or getattr(origin, "_name", None) == "Literal"
            ):  # Handle Literal types
                return value in args
            elif origin:  # Handle other generics like Dict, Tuple, etc.
                return isinstance(value, origin)
            else:  # Handle non-generic types
                return isinstance(value, hint)

        # Check types of arguments
        for arg_name, arg_value in args_with_hints.items():
            hint = hints.get(arg_name)
            if hint and not matches_type(arg_value, hint):
                raise ValueError(f"Argument {arg_name} does not match type {hint}; is {arg_value}")

        # Check types of keyword arguments
        for arg_name, arg_value in kwargs.items():
            hint = hints.get(arg_name)
            if hint and not matches_type(arg_value, hint):
                raise ValueError(f"Argument {arg_name} does not match type {hint}; is {arg_value} of type {type(arg_value)}")

        return func(*args, **kwargs)

    return wrapper


def annotate_message_json_list_with_tool_calls(messages: list[dict], allow_tool_roles: bool = False):
    """Add in missing tool_call_id fields to a list of messages using function call style

    Walk through the list forwards:
    - If we encounter an assistant message that calls a function ("function_call") but doesn't have a "tool_call_id" field
      - Generate the tool_call_id
    - Then check if the subsequent message is a role == "function" message
      - If so, then att
    """
    tool_call_index = None
    tool_call_id = None
    updated_messages = []

    for i, message in enumerate(messages):
        if "role" not in message:
            raise ValueError(f"message missing 'role' field:\n{message}")

        # If we find a function call w/o a tool call ID annotation, annotate it
        if message["role"] == "assistant" and "function_call" in message:
            if "tool_call_id" in message and message["tool_call_id"] is not None:
                printd("Message already has tool_call_id")
                tool_call_id = message["tool_call_id"]
            else:
                tool_call_id = str(uuid.uuid4())
                message["tool_call_id"] = tool_call_id
            tool_call_index = i

        # After annotating the call, we expect to find a follow-up response (also unannotated)
        elif message["role"] == "function":
            # We should have a new tool call id in the buffer
            if tool_call_id is None:
                # raise ValueError(
                print(
                    f"Got a function call role, but did not have a saved tool_call_id ready to use (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                )
                # allow a soft fail in this case
                message["tool_call_id"] = str(uuid.uuid4())
            elif "tool_call_id" in message:
                raise ValueError(
                    f"Got a function call role, but it already had a saved tool_call_id (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                )
            elif i != tool_call_index + 1:
                raise ValueError(
                    f"Got a function call role, saved tool_call_id came earlier than i-1 (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                )
            else:
                message["tool_call_id"] = tool_call_id
                tool_call_id = None  # wipe the buffer

        elif message["role"] == "assistant" and "tool_calls" in message and message["tool_calls"] is not None:
            if not allow_tool_roles:
                raise NotImplementedError(
                    f"tool_call_id annotation is meant for deprecated functions style, but got role 'assistant' with 'tool_calls' in message (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                )

            if len(message["tool_calls"]) != 1:
                raise NotImplementedError(
                    f"Got unexpected format for tool_calls inside assistant message (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                )

            assistant_tool_call = message["tool_calls"][0]
            if "id" in assistant_tool_call and assistant_tool_call["id"] is not None:
                printd("Message already has id (tool_call_id)")
                tool_call_id = assistant_tool_call["id"]
            else:
                tool_call_id = str(uuid.uuid4())
                message["tool_calls"][0]["id"] = tool_call_id
                # also just put it at the top level for ease-of-access
                # message["tool_call_id"] = tool_call_id
            tool_call_index = i

        elif message["role"] == "tool":
            if not allow_tool_roles:
                raise NotImplementedError(
                    f"tool_call_id annotation is meant for deprecated functions style, but got role 'tool' in message (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                )

            # if "tool_call_id" not in message or message["tool_call_id"] is None:
            # raise ValueError(f"Got a tool call role, but there's no tool_call_id:\n{messages[:i]}\n{message}")

            # We should have a new tool call id in the buffer
            if tool_call_id is None:
                # raise ValueError(
                print(
                    f"Got a tool call role, but did not have a saved tool_call_id ready to use (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                )
                # allow a soft fail in this case
                message["tool_call_id"] = str(uuid.uuid4())
            elif "tool_call_id" in message and message["tool_call_id"] is not None:
                if tool_call_id is not None and tool_call_id != message["tool_call_id"]:
                    # just wipe it
                    # raise ValueError(
                    #     f"Got a tool call role, but it already had a saved tool_call_id (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                    # )
                    message["tool_call_id"] = tool_call_id
                    tool_call_id = None  # wipe the buffer
                else:
                    tool_call_id = None
            elif i != tool_call_index + 1:
                raise ValueError(
                    f"Got a tool call role, saved tool_call_id came earlier than i-1 (i={i}, total={len(messages)}):\n{messages[:i]}\n{message}"
                )
            else:
                message["tool_call_id"] = tool_call_id
                tool_call_id = None  # wipe the buffer

        else:
            # eg role == 'user', nothing to do here
            pass

        updated_messages.append(copy.deepcopy(message))

    return updated_messages


def version_less_than(version_a: str, version_b: str) -> bool:
    """Compare versions to check if version_a is less than version_b."""
    # Regular expression to match version strings of the format int.int.int
    version_pattern = re.compile(r"^\d+\.\d+\.\d+$")

    # Assert that version strings match the required format
    if not version_pattern.match(version_a) or not version_pattern.match(version_b):
        raise ValueError("Version strings must be in the format 'int.int.int'")

    # Split the version strings into parts
    parts_a = [int(part) for part in version_a.split(".")]
    parts_b = [int(part) for part in version_b.split(".")]

    # Compare version parts
    return parts_a < parts_b


def create_random_username() -> str:
    """Generate a random username by combining an adjective and a noun."""
    adjective = random.choice(ADJECTIVE_BANK).capitalize()
    noun = random.choice(NOUN_BANK).capitalize()
    return adjective + noun


def verify_first_message_correctness(
    response: ChatCompletionResponse, require_send_message: bool = True, require_monologue: bool = False
) -> bool:
    """Can be used to enforce that the first message always uses send_message"""
    response_message = response.choices[0].message

    # First message should be a call to send_message with a non-empty content
    if (hasattr(response_message, "function_call") and response_message.function_call is not None) and (
        hasattr(response_message, "tool_calls") and response_message.tool_calls is not None
    ):
        printd(f"First message includes both function call AND tool call: {response_message}")
        return False
    elif hasattr(response_message, "function_call") and response_message.function_call is not None:
        function_call = response_message.function_call
    elif hasattr(response_message, "tool_calls") and response_message.tool_calls is not None:
        function_call = response_message.tool_calls[0].function
    else:
        printd(f"First message didn't include function call: {response_message}")
        return False

    function_name = function_call.name if function_call is not None else ""
    if require_send_message and function_name != "send_message" and function_name != "archival_memory_search":
        printd(f"First message function call wasn't send_message or archival_memory_search: {response_message}")
        return False

    if require_monologue and (not response_message.content or response_message.content is None or response_message.content == ""):
        printd(f"First message missing internal monologue: {response_message}")
        return False

    if response_message.content:
        ### Extras
        monologue = response_message.content

        def contains_special_characters(s):
            special_characters = '(){}[]"'
            return any(char in s for char in special_characters)

        if contains_special_characters(monologue):
            printd(f"First message internal monologue contained special characters: {response_message}")
            return False
        # if 'functions' in monologue or 'send_message' in monologue or 'inner thought' in monologue.lower():
        if "functions" in monologue or "send_message" in monologue:
            # Sometimes the syntax won't be correct and internal syntax will leak into message.context
            printd(f"First message internal monologue contained reserved words: {response_message}")
            return False

    return True


def is_valid_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


@contextmanager
def suppress_stdout():
    """Used to temporarily stop stdout (eg for the 'MockLLM' message)"""
    new_stdout = io.StringIO()
    old_stdout = sys.stdout
    sys.stdout = new_stdout
    try:
        yield
    finally:
        sys.stdout = old_stdout


def open_folder_in_explorer(folder_path):
    """
    Opens the specified folder in the system's native file explorer.

    :param folder_path: Absolute path to the folder to be opened.
    """
    if not os.path.exists(folder_path):
        raise ValueError(f"The specified folder {folder_path} does not exist.")

    # Determine the operating system
    os_name = platform.system()

    # Open the folder based on the operating system
    if os_name == "Windows":
        # Windows: use 'explorer' command
        subprocess.run(["explorer", folder_path], check=True)
    elif os_name == "Darwin":
        # macOS: use 'open' command
        subprocess.run(["open", folder_path], check=True)
    elif os_name == "Linux":
        # Linux: use 'xdg-open' command (works for most Linux distributions)
        subprocess.run(["xdg-open", folder_path], check=True)
    else:
        raise OSError(f"Unsupported operating system {os_name}.")


# Custom unpickler
class OpenAIBackcompatUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        if module == "openai.openai_object":
            from letta.openai_backcompat.openai_object import OpenAIObject

            return OpenAIObject
        return super().find_class(module, name)


# def count_tokens(s: str, model: str = "gpt-4") -> int:
#    try:
#        encoding = tiktoken.encoding_for_model(model)
#    except KeyError:
#        print("Falling back to cl100k base for token counting.")
#        encoding = tiktoken.get_encoding("cl100k_base")
#    return len(encoding.encode(s))


def printd(*args, **kwargs):
    if DEBUG:
        print(*args, **kwargs)


def united_diff(str1: str, str2: str) -> str:
    lines1 = str1.splitlines(True)
    lines2 = str2.splitlines(True)
    diff = difflib.unified_diff(lines1, lines2)
    return "".join(diff)


def parse_json(string) -> dict:
    """Parse JSON string into JSON with both json and demjson"""
    result = None
    try:
        result = json_loads(string)
        if not isinstance(result, dict):
            raise ValueError(f"JSON from string input ({string}) is not a dictionary (type {type(result)}): {result}")
        return result
    except Exception as e:
        print(f"Error parsing json with json package, falling back to demjson: {e}")

    try:
        result = demjson.decode(string)
        if not isinstance(result, dict):
            raise ValueError(f"JSON from string input ({string}) is not a dictionary (type {type(result)}): {result}")
        return result
    except demjson.JSONDecodeError as e:
        print(f"Error parsing json with demjson package (fatal): {e}")
        raise e


def validate_function_response(
    function_response: Any, return_char_limit: int, strict: bool = False, truncate: bool = True
) -> str | dict[str, Any]:
    """Check to make sure that a function used by Letta returned a valid response. Truncates to return_char_limit if necessary.

    This makes sure that we can coerce the function_response into a string or dict that meets our criteria. We handle some soft coercion.
    If strict is True, we raise a ValueError if function_response is not a string or None.

    Returns:
        str or dict: Validated response. Dicts are returned as-is to avoid double JSON encoding by package_function_response.
    """
    if isinstance(function_response, str):
        function_response_string = function_response

    elif function_response is None:
        function_response_string = "None"

    elif strict:
        raise ValueError(f"Strict mode violation. Function returned type: {type(function_response).__name__}")

    elif isinstance(function_response, dict):
        # For dicts, check if truncation is needed
        if truncate and return_char_limit:
            # Convert to JSON string to check size
            json_str = json_dumps(function_response)
            if len(json_str) > return_char_limit:
                # If truncation is needed, return truncated string
                logger.warning(f"function return was over limit ({len(json_str)} > {return_char_limit}) and was truncated")
                return f"{json_str[:return_char_limit]}... [NOTE: function output was truncated since it exceeded the character limit ({len(json_str)} > {return_char_limit})]"
        # Otherwise return dict as-is to avoid double JSON encoding
        # package_function_response will handle the final JSON serialization
        return function_response
    else:
        logger.debug(f"Function returned type {type(function_response).__name__}. Coercing to string.")
        function_response_string = str(function_response)

    # TODO we should change this to a max token limit that's variable based on tokens remaining (or context-window)
    if truncate and return_char_limit and len(function_response_string) > return_char_limit:
        logger.warning(f"function return was over limit ({len(function_response_string)} > {return_char_limit}) and was truncated")
        function_response_string = f"{function_response_string[:return_char_limit]}... [NOTE: function output was truncated since it exceeded the character limit ({len(function_response_string)} > {return_char_limit})]"

    return function_response_string


def list_agent_config_files(sort="last_modified"):
    """List all agent config files, ignoring dotfiles."""
    agent_dir = os.path.join(LETTA_DIR, "agents")
    files = os.listdir(agent_dir)

    # Remove dotfiles like .DS_Store
    files = [file for file in files if not file.startswith(".")]

    # Remove anything that's not a directory
    files = [file for file in files if os.path.isdir(os.path.join(agent_dir, file))]

    if sort is not None:
        if sort == "last_modified":
            # Sort the directories by last modified (most recent first)
            files.sort(key=lambda x: os.path.getmtime(os.path.join(agent_dir, x)), reverse=True)
        else:
            raise ValueError(f"Unrecognized sorting option {sort}")

    return files


def list_human_files():
    """List all humans files"""
    defaults_dir = os.path.join(letta.__path__[0], "humans", "examples")
    user_dir = os.path.join(LETTA_DIR, "humans")

    letta_defaults = os.listdir(defaults_dir)
    letta_defaults = [os.path.join(defaults_dir, f) for f in letta_defaults if f.endswith(".txt")]

    if os.path.exists(user_dir):
        user_added = os.listdir(user_dir)
        user_added = [os.path.join(user_dir, f) for f in user_added]
    else:
        user_added = []
    return letta_defaults + user_added


def list_persona_files():
    """List all personas files"""
    defaults_dir = os.path.join(letta.__path__[0], "personas", "examples")
    user_dir = os.path.join(LETTA_DIR, "personas")

    letta_defaults = os.listdir(defaults_dir)
    letta_defaults = [os.path.join(defaults_dir, f) for f in letta_defaults if f.endswith(".txt")]

    if os.path.exists(user_dir):
        user_added = os.listdir(user_dir)
        user_added = [os.path.join(user_dir, f) for f in user_added]
    else:
        user_added = []
    return letta_defaults + user_added


def get_human_text(name: str, enforce_limit=True):
    for file_path in list_human_files():
        file = os.path.basename(file_path)
        if f"{name}.txt" == file or name == file:
            with open(file_path, encoding="utf-8") as f:
                human_text = f.read().strip()
            if enforce_limit and len(human_text) > CORE_MEMORY_HUMAN_CHAR_LIMIT:
                raise ValueError(f"Contents of {name}.txt is over the character limit ({len(human_text)} > {CORE_MEMORY_HUMAN_CHAR_LIMIT})")
            return human_text

    raise ValueError(f"Human {name}.txt not found")


def get_persona_text(name: str, enforce_limit=True):
    for file_path in list_persona_files():
        file = os.path.basename(file_path)
        if f"{name}.txt" == file or name == file:
            with open(file_path, encoding="utf-8") as f:
                persona_text = f.read().strip()
            if enforce_limit and len(persona_text) > CORE_MEMORY_PERSONA_CHAR_LIMIT:
                raise ValueError(
                    f"Contents of {name}.txt is over the character limit ({len(persona_text)} > {CORE_MEMORY_PERSONA_CHAR_LIMIT})"
                )
            return persona_text

    raise ValueError(f"Persona {name}.txt not found")


def get_schema_diff(schema_a, schema_b):
    # Assuming f_schema and linked_function['json_schema'] are your JSON schemas
    f_schema_json = json_dumps(schema_a)
    linked_function_json = json_dumps(schema_b)

    # Compute the difference using difflib
    difference = list(difflib.ndiff(f_schema_json.splitlines(keepends=True), linked_function_json.splitlines(keepends=True)))

    # Filter out lines that don't represent changes
    difference = [line for line in difference if line.startswith("+ ") or line.startswith("- ")]

    return "".join(difference)


def create_uuid_from_string(val: str):
    """
    Generate consistent UUID from a string
    from: https://samos-it.com/posts/python-create-uuid-from-random-string-of-words.html
    """
    hex_string = hashlib.md5(val.encode("UTF-8")).hexdigest()
    return uuid.UUID(hex=hex_string)


def sanitize_filename(filename: str, add_uuid_suffix: bool = False) -> str:
    """
    Sanitize the given filename to prevent directory traversal, invalid characters,
    and reserved names while ensuring it fits within the maximum length allowed by the filesystem.

    Parameters:
        filename (str): The user-provided filename.
        add_uuid_suffix (bool): If True, adds a UUID suffix for uniqueness (legacy behavior).

    Returns:
        str: A sanitized filename.
    """
    # Extract the base filename to avoid directory components
    filename = os.path.basename(filename)

    # Split the base and extension
    base, ext = os.path.splitext(filename)

    # External sanitization library
    base = pathvalidate_sanitize_filename(base)

    # Cannot start with a period
    if base.startswith("."):
        raise ValueError(f"Invalid filename - derived file name {base} cannot start with '.'")

    if add_uuid_suffix:
        # Legacy behavior: Truncate the base name to fit within the maximum allowed length
        max_base_length = MAX_FILENAME_LENGTH - len(ext) - 33  # 32 for UUID + 1 for `_`
        if len(base) > max_base_length:
            base = base[:max_base_length]

        # Append a unique UUID suffix for uniqueness
        unique_suffix = uuid.uuid4().hex[:4]
        sanitized_filename = f"{base}_{unique_suffix}{ext}"
    else:
        max_base_length = MAX_FILENAME_LENGTH - len(ext)
        if len(base) > max_base_length:
            base = base[:max_base_length]

        sanitized_filename = f"{base}{ext}"

    # Return the sanitized filename
    return sanitized_filename


def get_friendly_error_msg(function_name: str, exception_name: str, exception_message: str):
    from letta.constants import MAX_ERROR_MESSAGE_CHAR_LIMIT

    error_msg = f"{ERROR_MESSAGE_PREFIX} executing function {function_name}: {exception_name}: {exception_message}"
    if len(error_msg) > MAX_ERROR_MESSAGE_CHAR_LIMIT:
        error_msg = error_msg[:MAX_ERROR_MESSAGE_CHAR_LIMIT]
    return error_msg


def parse_stderr_error_msg(stderr_txt: str, last_n_lines: int = 3) -> tuple[str, str]:
    """
    Parses out from the last `last_n_line` of `stderr_txt` the Exception type and message.
    """
    index = -(last_n_lines + 1)
    pattern = r"(\w+(?:Error|Exception)): (.+?)$"
    for line in stderr_txt.split("\n")[:index:-1]:
        if "Error" in line or "Exception" in line:
            match = re.search(pattern, line)
            if match:
                return match.group(1), match.group(2)
    return "", ""


def run_async_task(coro: Coroutine[Any, Any, Any]) -> Any:
    """
    Safely runs an asynchronous coroutine in a synchronous context.

    If an event loop is already running, it uses `asyncio.ensure_future`.
    Otherwise, it creates a new event loop and runs the coroutine.

    Args:
        coro: The coroutine to execute.

    Returns:
        The result of the coroutine.
    """
    try:
        # If there's already a running event loop, schedule the coroutine
        loop = asyncio.get_running_loop()
        return asyncio.run_until_complete(coro) if loop.is_closed() else asyncio.ensure_future(coro)
    except RuntimeError:
        # If no event loop is running, create a new one
        return asyncio.run(coro)


def log_telemetry(logger: Logger, event: str, **kwargs):
    """
    Logs telemetry events with a timestamp.

    :param logger: A logger
    :param event: A string describing the event.
    :param kwargs: Additional key-value pairs for logging metadata.
    """
    from letta.settings import log_settings

    if log_settings.verbose_telemetry_logging:
        timestamp = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S,%f UTC")  # More readable timestamp
        extra_data = " | ".join(f"{key}={value}" for key, value in kwargs.items() if value is not None)
        logger.info(f"[{timestamp}] EVENT: {event} | {extra_data}")


def make_key(*args, **kwargs):
    return str((args, tuple(sorted(kwargs.items()))))


# Global set to keep strong references to background tasks
_background_tasks: set = set()


def get_background_task_count() -> int:
    """Get the current number of background tasks for debugging/monitoring."""
    return len(_background_tasks)


@trace_method
def safe_create_task(coro, label: str = "background task"):
    # Check if coro is an async generator instead of a coroutine
    if inspect.isasyncgen(coro):
        raise TypeError(
            f"{label}: Cannot create task from async generator. "
            "Async generators must be consumed with 'async for', not 'await'. "
            "If you need to run an async generator as a task, wrap it in an async function."
        )

    if not inspect.iscoroutine(coro):
        raise TypeError(
            f"{label}: Expected a coroutine, got {type(coro).__name__}. Make sure you're calling the async function with () parentheses."
        )

    # Extract coroutine location for diagnostics
    coro_code = getattr(coro, "cr_code", None)
    if coro_code:
        filename = coro_code.co_filename
        idx = filename.find("letta/")
        filename = filename[idx + 6 :] if idx != -1 else filename.split("/")[-1]
        coro_name = f"{filename}:{coro_code.co_firstlineno}:{coro_code.co_name}"
    else:
        coro_name = "unknown"

    async def wrapper():
        try:
            await coro
        except Exception as e:
            # Don't log RunCancelledException as an error - it's expected when streams are cancelled
            if e.__class__.__name__ == "RunCancelledException":
                logger.info(f"{label} was cancelled (RunCancelledException)")
                return
            logger.exception(f"{label} failed with {type(e).__name__}: {e}")

    task = asyncio.create_task(wrapper(), name=f"safe[{coro_name}]")

    # Add task to the set to maintain strong reference
    _background_tasks.add(task)

    # Log task count to trace
    log_attributes({"total_background_task_count": get_background_task_count()})

    # Remove task from set when done to prevent memory leaks
    task.add_done_callback(_background_tasks.discard)

    return task


@trace_method
def safe_create_task_with_return(coro, label: str = "background task"):
    # Extract coroutine location for diagnostics
    coro_code = getattr(coro, "cr_code", None)
    if coro_code:
        filename = coro_code.co_filename
        idx = filename.find("letta/")
        filename = filename[idx + 6 :] if idx != -1 else filename.split("/")[-1]
        coro_name = f"{filename}:{coro_code.co_firstlineno}:{coro_code.co_name}"
    else:
        coro_name = "unknown"

    async def wrapper():
        try:
            return await coro
        except Exception as e:
            logger.exception(f"{label} failed with {type(e).__name__}: {e}")
            raise

    task = asyncio.create_task(wrapper(), name=f"safe_ret[{coro_name}]")

    # Add task to the set to maintain strong reference
    _background_tasks.add(task)

    # Log task count to trace
    log_attributes({"total_background_task_count": get_background_task_count()})

    # Remove task from set when done to prevent memory leaks
    task.add_done_callback(_background_tasks.discard)

    return task


def safe_create_shielded_task(coro, label: str = "shielded background task"):
    """
    Create a shielded background task that cannot be cancelled externally.

    This is useful for critical operations that must complete even if the
    parent operation is cancelled. The task is internally shielded but the
    returned task can still have callbacks added to it.
    """

    # Extract coroutine location for diagnostics
    coro_code = getattr(coro, "cr_code", None)
    if coro_code:
        filename = coro_code.co_filename
        idx = filename.find("letta/")
        filename = filename[idx + 6 :] if idx != -1 else filename.split("/")[-1]
        coro_name = f"{filename}:{coro_code.co_firstlineno}:{coro_code.co_name}"
    else:
        coro_name = "unknown"

    async def shielded_wrapper():
        try:
            # Shield the original coroutine to prevent cancellation
            result = await asyncio.shield(coro)
            return result
        except Exception as e:
            logger.exception(f"{label} failed with {type(e).__name__}: {e}")
            raise

    # Create the task with the shielded wrapper
    task = asyncio.create_task(shielded_wrapper(), name=f"safe_shield[{coro_name}]")

    # Add task to the set to maintain strong reference
    _background_tasks.add(task)

    # Log task count to trace
    log_attributes({"total_background_task_count": get_background_task_count()})

    # Remove task from set when done to prevent memory leaks
    task.add_done_callback(_background_tasks.discard)

    return task


def safe_create_file_processing_task(coro, file_metadata, server, actor, logger: Logger, label: str = "file processing task"):
    """
    Create a task for file processing that updates file status on failure.

    This is a specialized version of safe_create_task that ensures file
    status is properly updated to ERROR with a meaningful message if the
    task fails.

    Args:
        coro: The coroutine to execute
        file_metadata: FileMetadata object being processed
        server: Server instance with file_manager
        actor: User performing the operation
        logger: Logger instance for error logging
        label: Description of the task for logging
    """
    from letta.schemas.enums import FileProcessingStatus

    async def wrapper():
        try:
            await coro
        except Exception as e:
            logger.exception(f"{label} failed for file {file_metadata.file_name} with {type(e).__name__}: {e}")
            # update file status to ERROR with a meaningful message
            try:
                await server.file_manager.update_file_status(
                    file_id=file_metadata.id,
                    actor=actor,
                    processing_status=FileProcessingStatus.ERROR,
                    error_message=f"Processing failed: {str(e)}" if str(e) else f"Processing failed: {type(e).__name__}",
                )
            except Exception as update_error:
                logger.error(f"Failed to update file status to ERROR for {file_metadata.id}: {update_error}")

    task = asyncio.create_task(wrapper())

    # Add task to the set to maintain strong reference
    _background_tasks.add(task)

    # Remove task from set when done to prevent memory leaks
    task.add_done_callback(_background_tasks.discard)

    return task


async def get_latest_alembic_revision() -> str:
    """Get the current alembic revision ID from the alembic_version table."""
    from letta.server.db import db_registry

    try:
        async with db_registry.async_session() as session:
            result = await session.execute(text("SELECT version_num FROM alembic_version"))
            row = result.fetchone()

            if row:
                return row[0]
            else:
                return "unknown"

    except Exception as e:
        logger.error("Error getting latest alembic revision: %s", e)
        return "unknown"


def calculate_file_defaults_based_on_context_window(context_window: Optional[int]) -> tuple[int, int]:
    """Calculate reasonable defaults for max_files_open and per_file_view_window_char_limit
    based on the model's context window size.

    Args:
        context_window: The context window size of the model. If None, returns conservative defaults.

    Returns:
        A tuple of (max_files_open, per_file_view_window_char_limit)
    """
    if not context_window:
        # If no context window info, use conservative defaults
        return DEFAULT_MAX_FILES_OPEN, DEFAULT_CORE_MEMORY_SOURCE_CHAR_LIMIT

    # Define defaults based on context window ranges
    # Assuming ~4 chars per token
    # Available chars = available_tokens * 4

    # TODO: Check my math here
    if context_window <= 8_000:  # Small models (4K-8K)
        return 3, 5_000  # ~3.75K tokens
    elif context_window <= 32_000:  # Medium models (16K-32K)
        return 5, 15_000  # ~18.75K tokens
    elif context_window <= 128_000:  # Large models (100K-128K)
        return 10, 25_000  # ~62.5K tokens
    elif context_window <= 200_000:  # Very large models (128K-200K)
        return 10, 40_000  # ~100k tokens
    else:  # Extremely large models (200K+)
        return 15, 40_000  # ~1505k tokens


def truncate_file_visible_content(visible_content: str, is_open: bool, per_file_view_window_char_limit: int):
    visible_content = visible_content if visible_content and is_open else ""

    # Truncate content and add warnings here when converting from FileAgent to Block
    if len(visible_content) > per_file_view_window_char_limit:
        truncated_warning = f"...[TRUNCATED]\n{FILE_IS_TRUNCATED_WARNING}"
        visible_content = visible_content[: per_file_view_window_char_limit - len(truncated_warning)]
        visible_content += truncated_warning

    return visible_content


def fire_and_forget(coro, task_name: Optional[str] = None, error_callback: Optional[Callable[[Exception], None]] = None) -> asyncio.Task:
    """
    Execute an async coroutine in the background without waiting for completion.

    Args:
        coro: The coroutine to execute
        task_name: Optional name for logging purposes
        error_callback: Optional callback to execute if the task fails

    Returns:
        The created asyncio Task object
    """

    task = asyncio.create_task(coro)

    # Add task to the set to maintain strong reference
    _background_tasks.add(task)

    # Remove task from set when done to prevent memory leaks
    task.add_done_callback(_background_tasks.discard)

    def callback(t):
        try:
            t.result()  # this re-raises exceptions from the task
        except Exception as e:
            task_desc = f"Background task {task_name}" if task_name else "Background task"
            logger.exception(f"{task_desc} failed: {str(e)}")

            if error_callback:
                try:
                    error_callback(e)
                except Exception as callback_error:
                    logger.error(f"Error callback failed: {callback_error}")

    task.add_done_callback(callback)
    return task


def is_1_0_sdk_version(headers: HeaderParams):
    """
    Check if the SDK version is 1.0.0 or above.
    1. If sdk_version is provided from stainless (all stainless versions are 1.0.0+)
    2. If user_agent is provided and in the format
        @letta-ai/letta-client/version (node) or
        letta-client/version (python)
    """
    from letta.log import get_logger

    logger = get_logger(__name__)

    sdk_version = headers.sdk_version
    if sdk_version:
        return True

    client = headers.user_agent
    # None User-Agent might be like curl, which has no expectation of pinned behaviour etc.
    if not client:
        logger.debug("User-Agent header is None, assuming v1.0.0+")
        return True

    if "/" not in client:
        logger.debug(f"User-Agent '{client}' does not contain '/', assuming legacy SDK version")
        return False

    # Split into parts to validate format
    parts = client.split("/")

    # Should have at least 2 parts (client-name/version)
    if len(parts) < 2:
        return False

    if len(parts) == 3:
        # Format: @letta-ai/letta-client/version
        if parts[0] != "@letta-ai" or parts[1] != "letta-client":
            return False
    elif len(parts) == 2:
        # Format: letta-client/version
        if parts[0] != "letta-client":
            return False
    else:
        return False

    # Extract and validate version
    maybe_version = parts[-1]
    if "." not in maybe_version:
        return False

    # Extract major version (handle alpha/beta suffixes like 1.0.0-alpha.2 or 1.0.0a5)
    version_base = maybe_version.split("-")[0].split("a")[0].split("b")[0]
    if "." not in version_base:
        return False

    major_version = version_base.split(".")[0]
    return major_version == "1"


async def bounded_gather(coros: list[Coroutine], max_concurrency: int = 10) -> list:
    """
    Execute coroutines with bounded concurrency to prevent event loop saturation.

    Unlike asyncio.gather() which runs all coroutines concurrently, this limits
    the number of concurrent tasks to prevent overwhelming the event loop.

    Note: This is a stopgap measure. Prefer fixing the root cause by:
    - Limiting items fetched from DB (e.g., pagination)
    - Using explicit relationship loading instead of eager-loading all
    - Adding concurrency limits at the API/business logic layer

    Args:
        coros: List of coroutines to execute
        max_concurrency: Maximum number of concurrent tasks (default: 10)

    Returns:
        List of results in the same order as input coroutines
    """
    if not coros:
        return []

    semaphore = asyncio.Semaphore(max_concurrency)

    async def bounded_coro(index: int, coro: Coroutine, coro_name: str):
        # Set task name for diagnostics before acquiring semaphore
        task = asyncio.current_task()
        if task:
            task.set_name(f"bounded[{coro_name}]")

        async with semaphore:
            result = await coro
            return (index, result)

    # Wrap all coroutines with semaphore control, extracting location for diagnostics
    tasks = []
    for i, coro in enumerate(coros):
        coro_code = getattr(coro, "cr_code", None)
        if coro_code:
            filename = coro_code.co_filename
            idx = filename.find("letta/")
            filename = filename[idx + 6 :] if idx != -1 else filename.split("/")[-1]
            coro_name = f"{filename}:{coro_code.co_firstlineno}:{coro_code.co_name}"
        else:
            coro_name = "unknown"
        tasks.append(bounded_coro(i, coro, coro_name))
    indexed_results = await asyncio.gather(*tasks)

    # Sort by original index to preserve order
    indexed_results.sort(key=lambda x: x[0])
    return [result for _, result in indexed_results]


async def decrypt_agent_secrets(agents: list) -> list:
    """
    Decrypt secrets for all agents outside DB session.

    This allows DB connections to be released before expensive PBKDF2 operations,
    preventing connection pool exhaustion during high load.

    Uses bounded concurrency to limit thread pool pressure while allowing some
    parallelism in the dedicated crypto executor.

    Args:
        agents: List of PydanticAgentState objects with encrypted secrets

    Returns:
        Same list with secrets decrypted
    """

    async def decrypt_env_var(env_var):
        from letta.orm.agent import ENCRYPTED_PLACEHOLDER

        if env_var.value_enc and (env_var.value is None or env_var.value == "" or env_var.value == ENCRYPTED_PLACEHOLDER):
            env_var.value = await env_var.value_enc.get_plaintext_async()

    # Collect all env vars that need decryption
    decrypt_tasks = []
    for agent in agents:
        if agent.tool_exec_environment_variables:
            for env_var in agent.tool_exec_environment_variables:
                decrypt_tasks.append(decrypt_env_var(env_var))

    # Decrypt with bounded concurrency (matches crypto executor size)
    if decrypt_tasks:
        await bounded_gather(decrypt_tasks, max_concurrency=8)

    return agents



"""add_usage_columns_to_steps

Revision ID: 3e54e2fa2f7e
Revises: a1b2c3d4e5f8
Create Date: 2026-02-03 16:35:51.327031

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "3e54e2fa2f7e"
down_revision: Union[str, None] = "a1b2c3d4e5f8"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column("steps", sa.Column("model_handle", sa.String(), nullable=True))
    op.add_column("steps", sa.Column("cached_input_tokens", sa.Integer(), nullable=True))
    op.add_column("steps", sa.Column("cache_write_tokens", sa.Integer(), nullable=True))
    op.add_column("steps", sa.Column("reasoning_tokens", sa.Integer(), nullable=True))


def downgrade() -> None:
    op.drop_column("steps", "reasoning_tokens")
    op.drop_column("steps", "cache_write_tokens")
    op.drop_column("steps", "cached_input_tokens")
    op.drop_column("steps", "model_handle")
